"""Originally this script was used to update the checksums in CRDS mapping
files.   It has been extended so that it is capable of handling FITS reference
file checksums as well::

    % crds checksum  hst.pmap  hst_acs_0020.imap   hst_acs_darkfile_0057.rmap

    % crds checksum ./test.fits
    % crds checksum --remove ./*.fits
    % crds checksum --verify ./*.fits

no switch  adds / updates FITS or CRDS mapping checksums
--remove   removing FITS checksums  (mappings not supported)
--verify   verifying FITS or mapping checksums.

"""
import sys

# ============================================================================

from crds.core import log, rmap, cmdline, config, exceptions
from crds.certify import mapping_parser
from crds import data_file

# ============================================================================

def update_checksum(file_):
    """Rewrite the checksum of a single mapping, in place.   Reports duplicate cases being dropped."""
    if not check_duplicates(file_):
        mapping = rmap.Mapping.from_file(file_, ignore_checksum=True)
        mapping.write()

# Interim step to making update_checksums universal.  Switch to update_mapping_checksums now.
update_mapping_checksum = update_checksum

def check_duplicates(file_):
    """Before rewriting an rmap to update the checksum, certify to ensure no
    duplicates (or other certify errors) exist prior to rewriting checksum.
    """
    old_errs = log.errors()
    parsing = mapping_parser.parse_mapping(file_)
    mapping_parser.check_duplicates(parsing)
    new_errs = log.errors()
    return new_errs > old_errs

# ============================================================================

def add_checksum(file_):
    """Add checksums to file_."""
    log.info("Adding checksum for", repr(file_))
    if config.is_reference(file_):
        with log.error_on_exception("Failed updating checksum for", repr(file_)):
            data_file.add_checksum(file_)
    elif config.is_mapping(file_):
        update_mapping_checksum(file_)
    else:
        raise exceptions.CrdsError(
            "File", repr(file_), "does not appear to be a CRDS reference or mapping file.")

def remove_checksum(file_):
    """Remove checksums from `file_`."""
    log.info("Removing checksum for", repr(file_))
    if config.is_reference(file_):
        data_file.remove_checksum(file_)
    elif config.is_mapping(file_):
        raise exceptions.CrdsError("Mapping checksums cannot be removed for:", repr(file_))
    else:
        raise exceptions.CrdsError(
            "File", repr(file_), "does not appear to be a CRDS reference or mapping file.")

def verify_checksum(file_):
    """Verify checksums in `file_`."""
    log.info("Verifying checksum for", repr(file_))
    if config.is_reference(file_):
        data_file.verify_checksum(file_)
    elif config.is_mapping(file_):
        if config.CRDS_IGNORE_MAPPING_CHECKSUM.get():
            log.warning("Mapping checksums are disabled by config.CRDS_IGNORE_MAPPING_CHECKSM.")
        rmap.load_mapping(file_)
    else:
        raise exceptions.CrdsError(
            "File", repr(file_), "does not appear to be a CRDS reference or mapping file.")

# ============================================================================

class ChecksumScript(cmdline.Script):
    """Command line script for updating CRDS rule and reference file checksums."""

    description = """
    Add, remove, or verify checksums in CRDS rules or reference files.

    1. Default operation is to ADD checksums::

    % crds checksum  *.rmap

    % crds checksum  *.fits

    2. Reference files may support REMOVING checksums::

    % crds checksum --remove *.fits

    NOTE: CRDS mapping / rules files do not support removing checksums.

    3. Checksums can be VERIFIED without attempting to update or remove::

    % crds checksum --verify  *.rmap

    % crds checksum --verify *.fits

    Currently only FITS references support checksum operations.
    Checksums can be added or verified on all CRDS mapping types.
    """

    epilog = """
    """

    locate_file = cmdline.Script.locate_file_outside_cache

    def add_args(self):
        self.add_argument(
            "files", type=str, nargs="+",
            help="Files to operate on, CRDS rule or reference files.")

        self.add_argument(
            "--remove", action="store_true",
            help="Remove checksums when specified.  Invalid for CRDS mappings.")

        self.add_argument(
            "--verify", action="store_true",
            help="Verify checksums when specified.")

    def main(self):
        for file_ in self.files:
            with log.error_on_exception("Checksum operation FAILED"):
                if self.args.remove:
                    remove_checksum(file_)
                elif self.args.verify:
                    verify_checksum(file_)
                else:
                    add_checksum(file_)
        return log.errors()

if __name__ == "__main__":
    sys.exit(ChecksumScript()())



'''
This module defines the generic functions and abstract class used to manage
file access in CRDS.
'''
from collections.abc import Mapping
import functools
import warnings
import re

# ================================================================================================

import datetime
from astropy.time import Time

# ================================================================================================

# from astropy.utils.exceptions import AstropyUserWarning    # deferred

# ================================================================================================

from crds.core import exceptions, config, log, utils, timestamp

# ================================================================================================

DUPLICATES_OK = ["COMMENT", "HISTORY", "NAXIS","EXTNAME","EXTVER"]
APPEND_KEYS = ["COMMENT", "HISTORY"]

# ===========================================================================

# The point of hijack_warnings is to remap warnings from other packages to CRDS
# so that they are counted and logged and visible in web output.
# NOTE:  hijack_warnings needs to be nestable
# XXX: hijack_warnings is non-reentrant and FAILS with THREADS

def hijack_warnings(func):
    """Decorator that redirects warning messages to CRDS warnings."""
    @functools.wraps(func)
    def wrapper(*args, **keys):
        """Reassign warnings to CRDS warnings prior to executing `func`,  restore
        warnings state afterwards and return result of `func`.
        """
        # warnings.resetwarnings()
        from astropy.utils.exceptions import AstropyUserWarning
        with warnings.catch_warnings():
            old_showwarning = warnings.showwarning
            warnings.showwarning = hijacked_showwarning
            warnings.simplefilter("always", AstropyUserWarning)
            try:
                from stdatamodels.validate import ValidationWarning
            except:
                log.verbose_warning(
                    "stdatamodels ValidationWarning import failed.  "
                    "Not a problem for HST.",
                    verbosity=70)
            else:
                warnings.filterwarnings("always", r".*", ValidationWarning, r".*jwst.*")
                if not config.ALLOW_SCHEMA_VIOLATIONS:
                    warnings.filterwarnings("error", r".*is not one of.*", ValidationWarning, r".*jwst.*")
            try:
                result = func(*args, **keys)
            finally:
                warnings.showwarning = old_showwarning
        return result
    return wrapper

def hijacked_showwarning(message, category, filename, lineno, *args, **keys):
    """Map the warnings.showwarning plugin function parameters onto log.warning."""
    try:
        scat = str(category).split(".")[-1].split("'")[0]
    except Exception:
        scat = category
    try:
        sfile = str(filename).split(".egg")[-1].split("site-packages")[-1].replace("/",".").replace(".py", "")
        while sfile.startswith(("/",".")):
            sfile = sfile[1:]
    except Exception:
        sfile = filename
    message = str(message).replace("\n","")
    log.warning(scat, ":", sfile, ":", message)

# ----------------------------------------------------------------------------------------------
#
# Cross-strapping is used to support different synonyms for the same keyword:
#
#   FITS
#   JWST data model dotted path
#   Ad hoc names appearing in un-modeled reference files
#
# It's complicated because of inconsistent support for data model nomenclature both
# within and between projects;  it's currently a JWST-only concept but is implemented
# in a pluggable way.
#
# For "un-modeled" references or if the decision is made to drop datamodels use, each
# observatories locator module can define CROSS_STRAPPED_KEYWORDS.  See an example, jwst.
#
# For projects supporting the datamodels,  the schema is used to define the correspondence
# between data model dotted paths and the keyword of the underlying file format.
#
def cross_strap_header(header):
    """Set up keyword equivalencies in a copy of `header`.  Ensure both FITS
    and datamodel dotted path variants are defined for each keyword.
    Also add variations defined by observatory locator module
    CROSS_STRAPPED_KEYWORDS.
    """
    crossed = dict(header)
    try:
        locator = utils.header_to_locator(header)
    except Exception:
        log.verbose_warning(
            "Cannot identify observatory from header. Skipping keyword aliasing")
        return crossed
    equivalency_pairs = locator.get_cross_strapped_pairs(header)
    for pair in equivalency_pairs:
        _cross_strap_pair(crossed, pair)
    return crossed

def equivalence_dict_to_pairs(equivalent_keywords_dict):
    """Convert a dictionary mapping master keywords to equivalents to
    a list of keyword pairs that should be cross-strapped.
    """
    pairs = []
    log.verbose("Explicitly cross_strapped_keywords:",
                log.PP(equivalent_keywords_dict), verbosity=90)
    for master, slaves in equivalent_keywords_dict.items():
        for slave in slaves:
            if master != slave:
                pairs.append((master, slave))
                pairs.append((slave, master))
    return pairs


def _cross_strap_pair(header, keyword_pair):
    """Mutate `header` using (master, slave) `keyword_pair` so that slave
    duplicates master's value under slave's name IFF master is defined
    in header and slave is not.
    """
    master_key, slave_key = keyword_pair
    master_val = header.get(master_key, "UNDEFINED")
    slave_val =  header.get(slave_key, "UNDEFINED")
    if slave_val == "UNDEFINED" and master_val != "UNDEFINED":
        header[slave_key] = master_val

# ----------------------------------------------------------------------------------------------

def convert_to_eval_header(header):
    """To support using file headers in eval expressions,  two changes need to be made:

    1. JWST data model keywords, dotted paths, need to be translated to underscored paths.
    This makes them valid identifiers instead of non-existent nested objects when evaled.

    2. Numerial values that CRDS has conditioned into strings need to be converted back to
    numerical values so they can be used in arithmetic expressions.
    """
    header = _destringize_numbers(header)
    header = _convert_dotted_paths(header)
    return header

def _destringize_numbers(header):
    """Convert string values in `header` that work as numbers back into
    ints and floats.
    """
    with_numbers = {}
    for key, val in header.items():
        try:
            val = int(val)
        except:
            try:
                val = float(val)
            except:
                pass
        with_numbers[key] = val
    return with_numbers

def _convert_dotted_paths(header):
    """Convert header dotted-path keys into valid Python identifiers
    (for eval()) by using underscores instead of periods and add to
    existing contents of `header`.
    """
    cleaned = dict(header)
    for key, val in header.items():
        clean = re.sub(r"([A-Za-z][A-Za-z0-9_]*)\.", r"\1_", key)
        cleaned[clean] = val
    return cleaned

# ----------------------------------------------------------------------------------------------

def ensure_keys_defined(header, needed_keys=(), define_as="UNDEFINED"):
    """Define any keywords from `needed_keys` which are missing in `header`,  or defined as 'UNDEFINED',
    as `default`.

    Normally this defines missing keys as UNDEFINED.

    It can be used to redefine UNDEFINED as something else,  like N/A.
    """
    header = dict(header)
    for key in needed_keys:
        if key not in header or header[key] in ["UNDEFINED", None]:
            header[key] = define_as
    return header

# ================================================================================================

class AbstractFile:

    format = "ABSTRACT"

    def __init__(self, filepath, original_name=None, observatory=None):
        """Create an AbstractFile.

        Required for use:

        filepath          working path for this file

        Retained for debug:

        original_name     abstract,  possibly more readable name to infer type and observatory w/o io
        observatory       observatory this file belongs to
        """
        # This __init__ should be kept fast
        self.filepath = filepath
        self.original_name = original_name
        self.observatory = observatory
        self.array_formats = {}

    def _unsupported_file_op_error(self, method):
        return exceptions.UnsupportedFileOpError(
            "Method", repr(method), "is not supported for file format", repr(self.format))

    def add_checksum(self):
        """Add checksum to`self.filepath`."""
        raise self._unsupported_file_op_error("add_checksum")

    def remove_checksum(self):
        """Remove checksum from`self.filepath`."""
        raise self._unsupported_file_op_error("remove_checksum")

    def verify_checksum(self):
        """Verify checksum in `self.filepath`."""
        raise self._unsupported_file_op_error("verify_checksum")

    def get_format(self):
        """Return a string describing the structure of file at `filepath`,  intended
        for file overview describing generic array structure.
        """
        raise self._unsupported_file_op_error("get_format")

    def get_array_properties(self, array_name, keytype="A"):
        """Return a basic properties dictionary for array named `array_name`."""
        raise self._unsupported_file_op_error("get_array_properties")

    def get_array(self, array_name):
        """Return the array object corresponding to array selected by `array_id_info`."""
        raise self._unsupported_file_op_error("get_array")

    # ----------------------------------------------------------------------------------------------

    def getval(self, key, **keys):
        """Return a single metadata value from `key` of file at `filepath`."""
        return self.get_header((key,), **keys)[key]

    def setval(self, key, value):
        """Set the value of a single metadata key,  nominally in the 'primary header'."""
        raise self._unsupported_file_op_error("setval")

    # ----------------------------------------------------------------------------------------------

    def get_header(self, needed_keys, **keys):
        """Return dictionary of metadata for this file,  e.g. FITS primary header
         dictionary featuring keywords `needed_keys`.
         """
        raw_header = self.get_raw_header(needed_keys, **keys)
        reduced_header = self._reduce_header(raw_header, needed_keys)
        crossed_header = cross_strap_header(reduced_header)
        crossed_header["FILE_FORMAT"] = \
            self.__class__.__name__[:-len("File")].upper()
        return crossed_header

    def get_raw_header(self, needed_keys, **keys):
        """Return the metadata dictionary associated with this file,  nominally a dict
        describing the FITS header, ASDF tree, or JSON or YAML contents.
        """
        raise self._unsupported_file_op_error("get_raw_header")
    # ----------------------------------------------------------------------------------------------

    def _reduce_header(self, old_header, needed_keys=()):
        """Limit `header` to `needed_keys`,  converting all keys to upper case
        and making note of any significant duplicates, and adding any missing
        `needed_keys` as UNDEFINED.

        To detect duplicates,  use an item list for `old_header`,  not a dict.
        """
        needed_keys = tuple(key.upper() for key in needed_keys)
        header = {}
        if isinstance(old_header, dict):
            old_header = old_header.items()
        for (key, value) in old_header:
            key = str(key).upper()
            value = str(value)
            if (not needed_keys) or key in needed_keys:
                if key in header and header[key] != value:
                    if key not in DUPLICATES_OK:
                        log.verbose_warning("Duplicate key", repr(key), "in", repr(self.filepath),
                                            "using", repr(header[key]), "not", repr(value), verbosity=70)
                        continue
                    elif key in APPEND_KEYS:
                        header[key] += "\n" + value
                else:
                    header[key] = value
        return ensure_keys_defined(header, needed_keys)

    # ----------------------------------------------------------------------------------------------

    def to_simple_types(self, tree):
        """Convert a tree structure to a flat dictionary of simple types with dotted path tree keys."""
        result = dict()
        for key in tree:
            if not isinstance(key, str):  # skip non-string keys
                continue
            value = tree[key]
            if isinstance(value, Mapping):
                nested = self.to_simple_types(value)
                for nested_key, nested_value in nested.items():
                    result[str(key.upper() + "." + nested_key)] = nested_value
            else:
                result[str(key.upper())] = self._simple_type(value)
        return result

    def _simple_type(self, value):
        """Convert ASDF values to simple strings, where applicable,  exempting potentially large values."""
        if isinstance(value, (str, int, float, complex)):
            rval = str(value)
        elif isinstance(value, (list, tuple)):
            rval = tuple(self._simple_type(val) for val in value)
        elif isinstance(value, (datetime.datetime, Time)):
            rval = timestamp.reformat_date(value).replace(" ", "T")
        else:
            rval = "SUPRESSED_NONSTD_TYPE: " + repr(str(value.__class__.__name__))
        return rval

    def get_asdf_standard_version(self):
        """
        Return the ASDF Standard version associated with this file as a string,
        or `None` if the file is neither an ASDF file nor contains an embedded
        ASDF file.
        """
        return None



"""This module contains doctests and unit tests which exercise the module that
builds on the crds.client web api and enhances it with fully developed cache
semantics.  crds.client is largely decoupled from the core CRDS modules and
focuses on JSONRPC calls and file downloads.   The heavy_client extends those
primitive functions to provide higher level APIs that are dependent on the
core library to implement.
"""
import os
import json
from pprint import pprint as pp
import doctest


from crds.core import rmap, log, config, heavy_client
from crds.core.exceptions import *
from crds import tests
from crds.tests import test_config
from crds.client import api

from nose.tools import assert_raises, assert_true

# ==================================================================================

def dt_getreferences_rmap_na():
    """
    >>> old_state = test_config.setup(cache=None, url="https://jwst-crds.stsci.edu")
    >>> os.environ["CRDS_MAPPATH_SINGLE"] = test_config.TEST_DATA

    >>> heavy_client.getreferences({"META.INSTRUMENT.NAME":"NIRISS", "META.INSTRUMENT.DETECTOR":"NIS",
    ...                             "META.INSTRUMENT.FILTER":"BOGUS2", "META.EXPOSURE.TYPE":"NIS_IMAGE"},
    ...    observatory="jwst", context="jwst_na_omit.pmap", ignore_cache=False, reftypes=["flat"])
    {'flat': 'NOT FOUND n/a'}

    >>> test_config.cleanup(old_state)

    >> config.get_crds_state()
    """

def dt_getreferences_rmap_omit():
    """
    >>> old_state = test_config.setup(cache=None, url="https://jwst-crds.stsci.edu")
    >>> os.environ["CRDS_MAPPATH_SINGLE"] = test_config.TEST_DATA

    >>> heavy_client.getreferences({"META.INSTRUMENT.NAME":"NIRISS", "META.INSTRUMENT.DETECTOR":"NIS", "META.INSTRUMENT.FILTER":"BOGUS1", "META.EXPOSURE.TYPE":"NIS_IMAGE"},
    ...    observatory="jwst", context="jwst_na_omit.pmap", ignore_cache=False, reftypes=["flat"])
    {}

    >>> test_config.cleanup(old_state)

    >> config.get_crds_state()
    """

def dt_getreferences_imap_na():
    """
    >>> old_state = test_config.setup(cache=None, url="https://jwst-crds.stsci.edu")
    >>> os.environ["CRDS_MAPPATH_SINGLE"] = test_config.TEST_DATA

    >>> heavy_client.getreferences({"META.INSTRUMENT.NAME":"FGS", "META.EXPOSURE.TYPE":"FGS_IMAGE"},
    ...    observatory="jwst", context="jwst_na_omit.pmap", ignore_cache=False, reftypes=["flat"])
    {'flat': 'NOT FOUND n/a'}

    >>> test_config.cleanup(old_state)
    """

def dt_getreferences_imap_omit():
    """
    >>> old_state = test_config.setup(cache=None, url="https://jwst-crds.stsci.edu")
    >>> os.environ["CRDS_MAPPATH_SINGLE"] = test_config.TEST_DATA

    >>> heavy_client.getreferences({"META.INSTRUMENT.NAME":"MIRI", "META.EXPOSURE.TYPE":"MIR_IMAGE"},
    ...    observatory="jwst", context="jwst_na_omit.pmap", ignore_cache=False, reftypes=["flat"])
    {}

    >>> test_config.cleanup(old_state)
    """

def dt_getreferences_ignore_cache():
    """
    >>> old_state = test_config.setup(url="https://jwst-crds.stsci.edu")
    >>> header = { "META.INSTRUMENT.NAME":"MIRI", "META.EXPOSURE.TYPE":"MIR_IMAGE",
    ...            "META.OBSERVATION.DATE":"2018-05-25", "META.OBSERVATION.TIME":"00:00:00" }
    >>> heavy_client.getreferences(header,
    ...   observatory="jwst", context="jwst_miri.imap", ignore_cache=True, reftypes=["flat"])   # doctest: +ELLIPSIS
    CRDS - INFO -  Fetching  .../crds-cache-default-test/mappings/jwst/jwst_miri_photom.rmap      399 bytes  (1 / 3 files) (0 / 1.2 K bytes)
    CRDS - INFO -  Fetching  .../crds-cache-default-test/mappings/jwst/jwst_miri_flat.rmap      427 bytes  (2 / 3 files) (399 / 1.2 K bytes)
    CRDS - INFO -  Fetching  .../crds-cache-default-test/mappings/jwst/jwst_miri.imap      358 bytes  (3 / 3 files) (826 / 1.2 K bytes)
    CRDS - INFO -  Fetching  .../crds-cache-default-test/references/jwst/jwst_miri_flat_0001.fits   42.0 M bytes  (1 / 1 files) (0 / 42.0 M bytes)
    {'flat': '.../crds-cache-default-test/references/jwst/jwst_miri_flat_0001.fits'}
    >>> test_config.cleanup(old_state)
    """

def dt_cache_references_multiple_bad_files():
    """

    Define bestrefs with multiple errors which should all be reported
    prior to raising an exception.

    >>> old_state = test_config.setup()

    >>> bestrefs = {
    ...    "flat": "NOT FOUND something went wrong for flat.",
    ...    "dark": "NOT FOUND something else went wrong for dark.",
    ... }

    To work effectively for JWST reference file coverage improvement,
    CRDS needs to report ALL bad references in a given Step or
    prefetch, not just raise an exception on the first bad reference
    found.

    >>> api.cache_references("jwst.pmap", bestrefs)
    Traceback (most recent call last):
    ...
    CrdsLookupError: Error determining best reference for 'flat'  =   something went wrong for flat.

    >>> test_config.cleanup(old_state)

    """

def dt_get_context_name_literal():
    """
    >>> old_state = test_config.setup(url="https://jwst-crds-serverless.stsci.edu", observatory="jwst")
    >>> heavy_client.get_context_name("jwst", "jwst_0341.pmap")
    'jwst_0341.pmap'
    >>> test_config.cleanup(old_state)
    """

def dt_get_context_name_crds_context():
    """
    >>> old_state = test_config.setup(url="https://jwst-crds-serverless.stsci.edu", observatory="jwst")
    >>> os.environ["CRDS_CONTEXT"] = "jwst_0399.pmap"
    >>> heavy_client.get_context_name("jwst")
    'jwst_0399.pmap'
    >>> del os.environ["CRDS_CONTEXT"]
    >>> test_config.cleanup(old_state)
    """

def dt_get_context_name_symbolic():
    """
    >>> old_state = test_config.setup(url="https://jwst-crds-serverless.stsci.edu", observatory="jwst")
    >>> heavy_client.get_context_name("jwst", "jwst-operational")  # doctest: +ELLIPSIS
    'jwst_...pmap'
    >>> heavy_client.get_context_name("jwst", "jwst-edit")  # doctest: +ELLIPSIS
    'jwst_...pmap'
    >>> heavy_client.get_context_name("jwst", "jwst-versions")  # doctest: +ELLIPSIS
    'jwst_...pmap'
    >>> test_config.cleanup(old_state)
    """

def dt_translate_date_based_context():
    """
    >>> old_state = test_config.setup(url="https://jwst-crds-serverless.stsci.edu", observatory="jwst")
    >>> heavy_client.translate_date_based_context("foo-edit", observatory=None)
    Traceback (most recent call last):
    ...
    crds.core.exceptions.CrdsError: Cannot determine observatory to translate mapping 'foo-edit'
    >>> heavy_client.translate_date_based_context("jwst-2018-01-01T00:00:00")
    Traceback (most recent call last):
    ...
    crds.core.exceptions.CrdsError: Specified CRDS context by date 'jwst-2018-01-01T00:00:00' and CRDS server is not reachable.
    >>> test_config.cleanup(old_state)

    >>> old_state = test_config.setup(url="https://jwst-crds.stsci.edu", observatory="jwst")
    >>> heavy_client.translate_date_based_context("jwst-foo-2018-01-01T00:00:00")
    Traceback (most recent call last):
    ...
    crds.core.exceptions.ServiceError: CRDS jsonrpc failure 'get_context_by_date' InvalidDateBasedContext: Bad instrument 'foo' in CRDS date based context specification.
    >>> test_config.cleanup(old_state)
    """
def dt_get_bad_mappings_in_context_no_instrument():
    """
    >>> old_state = test_config.setup(url="https://jwst-crds-serverless.stsci.edu", observatory="jwst")
    >>> heavy_client.get_bad_mappings_in_context("jwst", "jwst_0016.pmap")
    ['jwst_miri_flat_0002.rmap']
    >>> test_config.cleanup(old_state)
    """

def dt_pickled_mappings(mapping):
    """
    >>> old_state = test_config.setup()

    >>> pickle_file = config.locate_pickle("jwst_0016.pmap","jwst")
    >>> pickle_file   # doctest: +ELLIPSIS
    '.../pickles/jwst/jwst_0016.pmap.pkl'

    >>> _ = heavy_client.get_pickled_mapping("jwst_0016.pmap", cached=True, use_pickles=True, save_pickles=True)  # doctest: +ELLIPSIS
    CRDS - INFO -  Saved pickled context '.../crds-cache-default-test/pickles/jwst/jwst_0016.pmap.pkl'
    >>> assert os.path.exists(pickle_file)

    >>> _ = heavy_client.get_pickled_mapping("jwst_0016.pmap", cached=True, use_pickles=True, save_pickles=False)  # doctest: +ELLIPSIS
     CRDS - INFO -  Loaded pickled context 'jwst_0016.pmap'

    >>> _ = heavy_client.get_pickled_mapping("jwst_0016.pmap", cached=False, use_pickles=False, save_pickles=False)  # doctest: +ELLIPSIS

    >>> _ = heavy_client.get_pickled_mapping("jwst_0016.pmap", cached=False, use_pickles=True, save_pickles=False)  # doctest: +ELLIPSIS
     CRDS - INFO -  Loaded pickled context 'jwst_0016.pmap'

    >>> heavy_client.load_pickled_mapping("jwst_0016.pmap")
    CRDS - INFO -  Loaded pickled context 'jwst_0016.pmap'
    PipelineContext('jwst_0016.pmap')

    >>> _ = log.set_verbose()
    >>> os.chmod(pickle_file, 0o444)
    >>> heavy_client.remove_pickled_mapping("jwst_0016.pmap")  # doctest: +ELLIPSIS
    CRDS - DEBUG -  Pickle file '.../crds-cache-default-test/pickles/jwst/jwst_0016.pmap.pkl' is not writable,  skipping pickle remove.

    >>> os.chmod(pickle_file, 0o666)
    >>> heavy_client.remove_pickled_mapping("jwst_0016.pmap")  # doctest: +ELLIPSIS
    CRDS - INFO -  Removed pickle for context '.../pickles/jwst/jwst_0016.pmap.pkl'
    >>> assert not os.path.exists(pickle_file)

    >>> heavy_client.remove_pickled_mapping("somewhere/foo.pmap")  # doctest: +ELLIPSIS
    CRDS - DEBUG -  Pickle file 'somewhere/foo.pmap' is not writable,  skipping pickle remove.

    >>> test_config.cleanup(old_state)
    """

def dt_check_parameters():
    """
    >>> old_state = test_config.setup(url="https://jwst-crds-serverless.stsci.edu", observatory="jwst")

    >>> header = { "NAME" : "VALID_VALUE",  "NAME1" : 1.0, "META.VALID.NAME3" : 1,   "NAME4" : True}
    >>> heavy_client.check_parameters(header)
    {'NAME': 'VALID_VALUE', 'NAME1': 1.0, 'META.VALID.NAME3': 1, 'NAME4': True}

    >>> header = { (1,2,3) : "something for invalid key" }
    >>> heavy_client.check_parameters(header)
    Traceback (most recent call last):
    ...
    AssertionError: Non-string key (1, 2, 3) in parameters.

    >>> header = { "META.BAD.VALUE" : object() }
    >>> heavy_client.check_parameters(header)
    {}

    >>> test_config.cleanup(old_state)
    """

def dt_check_context():
    """
    >>> heavy_client.check_context(None)
    """

def dt_get_context_parkeys():
    """
    >>> old_state = test_config.setup(url="https://jwst-crds-serverless.stsci.edu", observatory="jwst")
    >>> heavy_client.get_context_parkeys("jwst.pmap","miri")
    ['META.INSTRUMENT.TYPE', 'META.INSTRUMENT.LAMP_STATE', 'META.OBSERVATION.DATE', 'META.VISIT.TSOVISIT', 'REFTYPE']
    >>> heavy_client.get_context_parkeys("jwst_miri.imap","miri")
    ['META.INSTRUMENT.LAMP_STATE', 'META.OBSERVATION.DATE', 'META.VISIT.TSOVISIT', 'REFTYPE']
    >>> heavy_client.get_context_parkeys("jwst_miri_flat.rmap","miri")
    ['META.OBSERVATION.DATE', 'META.VISIT.TSOVISIT', 'META.INSTRUMENT.LAMP_STATE']
    >>> test_config.cleanup(old_state)
    """

# ==================================================================================

# class TestHeavyClient(test_config.CRDSTestCase):

    """
    def test_rmap_get_imap_except(self):
        r = rmap.get_cached_mapping("hst.pmap")
        with self.assertRaises(CrdsUnknownInstrumentError):
            r.get_imap("foo")
    """

# ==================================================================================

def tst():
    """Run module tests,  for now just doctests only."""
    # import unittest
    # suite = unittest.TestLoader().loadTestsFromTestCase(TestHeavyClient)
    # unittest.TextTestRunner().run(suite)

    from crds.tests import test_heavy_client, tstmod
    return tstmod(test_heavy_client)

if __name__ == "__main__":
    print(tst())



"""This module defines the API for CRDS clients.   Functions defined here make
remote service calls to the CRDS server to obtain mapping or reference files and
cache them locally.
"""
import os
import os.path
import base64
import re
import zlib
import html
from urllib import request
import warnings
import json
import ast

# ==============================================================================

# heavy versions of core CRDS modules defined in one place, client minimally
# dependent on core for configuration, logging, and  file path management.
# import crds
from crds.core import utils, log, config, constants
from crds.core.log import srepr

from crds.core.exceptions import ServiceError, CrdsLookupError
from crds.core.exceptions import CrdsNetworkError, CrdsDownloadError
from crds.core.exceptions import CrdsRemoteContextError

from . import proxy
from .proxy import CheckingProxy

# ==============================================================================

__all__ = [
    "get_default_observatory",
    "get_default_context",
    "get_context_by_date",
    "get_server_info",
    "get_cached_server_info",  # deprecated
    "cache_references",

    "set_crds_server",
    "get_crds_server",

    "list_mappings",
    "list_references",

    "get_url",      # deprecated
    "get_flex_uri",
    "get_file_info",
    "get_file_info_map",
    "get_sqlite_db",

    "get_mapping_names",
    "get_reference_names",

    "dump_references",
    "dump_mappings",
    "dump_files",

    "get_best_references",
    "cache_best_references",

    "get_dataset_headers_by_id",
    "get_dataset_headers_by_instrument",
    "get_dataset_ids",
    "get_best_references_by_ids",
    "get_aui_best_references",
    "get_best_references_by_header_map",

    "get_required_parkeys",
    "get_affected_datasets",
    "get_context_history",

    "push_remote_context",
    "get_remote_context",

    "jpoll_pull_messages",
    "jpoll_abort",

    "get_system_versions",
    ]

# ============================================================================

# Server for CRDS services and mappings

URL_SUFFIX = "/json/"

S = None    # Proxy server

def set_crds_server(url):
    """Configure the CRDS JSON services server to `url`,
    e.g. 'http://localhost:8000'
    """
    if not url.startswith("https://") and "localhost" not in url:
        log.warning("CRDS_SERVER_URL does not start with https://  ::", url)
    if url.endswith("/"):
        url = url[:-1]
    global URL, S
    URL = url + URL_SUFFIX
    S = CheckingProxy(URL, version="1.0")

def get_crds_server():
    """Return the base URL for the CRDS JSON RPC server.
    """
    url = URL[:-len(URL_SUFFIX)]
    if not url.startswith("https://") and "localhost" not in url:
        log.warning("CRDS_SERVER_URL does not start with https://  ::", url)
    return url

# =============================================================================

@utils.cached
def list_mappings(observatory=None, glob_pattern="*"):
    """Return the list of mappings associated with `observatory`
    which match `glob_pattern`.
    """
    return [str(x) for x in S.list_mappings(observatory, glob_pattern)]

@utils.cached
def list_references(observatory=None, glob_pattern="*"):
    """Return the list of references associated with `observatory`
    which match `glob_pattern`.
    """
    return [str(x) for x in S.list_references(observatory, glob_pattern)]

def get_mapping_url(pipeline_context, mapping):
    """Returns a URL for the specified pmap, imap, or rmap file.   DEPRECATED
    """
    utils.deprecate(
        "crds.client.get_mapping_url()", "2020-09-01", "crds.client.get_flex_uri()")
    return S.get_mapping_url(pipeline_context, mapping)

def is_known_mapping(mapping):
    """Return True iff `mapping` is a known/official CRDS mapping file."""
    try:
        return len(get_mapping_url(mapping, mapping)) > 0
    except ServiceError:
        return False

@utils.cached
def get_mapping_names(pipeline_context):
    """Get the complete set of pmap, imap, and rmap basenames required
    for the specified pipeline_context.   context can be an observatory,
    pipeline, or instrument context.
    """
    return [str(x) for x in S.get_mapping_names(pipeline_context)]

def get_reference_url(pipeline_context, reference):
    """Returns a URL for the specified reference file.    DEPRECATED
    """
    utils.deprecate(
        "crds.client.get_reference_url()", "2020-09-01", "crds.client.get_flex_uri()")
    return S.get_reference_url(pipeline_context, reference)

def get_url(pipeline_context, filename):
    """Return the URL for a CRDS reference or mapping file.   DEPRECATED
    """
    utils.deprecate(
        "crds.client.get_url()", "2020-09-01", "crds.client.get_flex_uri()")
    return S.get_url(pipeline_context, filename)

def get_flex_uri(filename, observatory=None):
    """If environment variables define the base URI for `filename`, append
    filename and return the combined URI.

    If no environment override has been specified, obtain the base URI from
    the server_info config,  append filename, and return the combined URI.

    If `filename` is a config file and no environment override is defined,
    return "none".
    """
    if observatory is None:
        observatory = get_default_observatory()
    uri = config.get_uri(filename)
    if uri == "none":
        info = get_server_info()
        if config.is_config(filename):
            uri = _unpack_info(info, "config_url", observatory)
        elif config.is_pickle(filename):
            uri = _unpack_info(info, "pickle_url", observatory)
        elif config.is_mapping(filename):
            uri = _unpack_info(info, "mapping_url", observatory)
        elif config.is_reference(filename):
            uri = _unpack_info(info, "reference_url", observatory)
        else:
            raise CrdsError("Can't identify file type for:", srepr(filename))
        if uri == "none":
            return uri
        if not uri.endswith("/"):
            uri += "/"
        uri += filename
    return uri

def _unpack_info(info, section, observatory):
    """Return info[section][observatory] if info[section] is defined.
    Otherwise return "none".
    """
    sect = info.get(section)
    if sect:
       return sect[observatory]
    else:
        return "none"

def get_file_info(pipeline_context, filename):
    """Return a dictionary of CRDS information about `filename`."""
    return S.get_file_info(pipeline_context, filename)

def get_file_info_map(observatory, files=None, fields=None):
    """Return the info { filename : { info } } on `files` of `observatory`.
    `fields` can be used to limit info returned to specified keys.
    """
    if files is not None:
        files = tuple(sorted(files))
    if fields is not None:
        fields = tuple(sorted(fields))
    return _get_file_info_map(observatory, files, fields)

@utils.cached
def _get_file_info_map(observatory, files, fields):
    """Memory cached version of get_file_info_map() service."""
    infos = S.get_file_info_map(observatory, files, fields)
    return infos

def get_total_bytes(info_map):
    """Return the total byte count of file info map `info_map`."""
    try:
        return sum([int(info_map[name]["size"]) for name in info_map if "NOT FOUND" not in info_map[name]])
    except Exception as exc:
        log.error("Error computing total byte count: ", str(exc))
        return -1

def get_sqlite_db(observatory):
    """Download the CRDS database as a SQLite database."""
    assert not config.get_cache_readonly(), "Readonly cache, updating the SQLite database cannot be done."""
    encoded_compressed_data = S.get_sqlite_db(observatory)
    data = zlib.decompress(base64.b64decode(encoded_compressed_data))
    path = config.get_sqlite3_db_path(observatory)
    utils.ensure_dir_exists(path)
    with open(path, "wb+") as db_out:
        db_out.write(data)
    return path

@utils.cached
def get_reference_names(pipeline_context):
    """Get the complete set of reference file basenames required
    for the specified pipeline_context.
    """
    return [str(x) for x in S.get_reference_names(pipeline_context)]

def get_best_references(pipeline_context, header, reftypes=None):
    """Get best references for dict-like `header` relative to
    `pipeline_context`.

    pipeline_context  CRDS context for lookup,   e.g.   'hst_0001.pmap'
    header            dict-like mapping { lookup_parameter : value }
    reftypes         If None,  return all reference types;  otherwise return
                     best refs for the specified list of reftypes.

    Returns          { reftype : reference_basename ... }

    Raises           CrdsLookupError,  typically for problems with header values
    """
    header = { str(key):str(value) for (key,value) in header.items() }
    try:
        bestrefs = S.get_best_references(pipeline_context, dict(header), reftypes)
    except Exception as exc:
        raise CrdsLookupError(str(exc)) from exc
    return bestrefs

def get_best_references_by_ids(context, dataset_ids, reftypes=None, include_headers=False):
    """Get best references for the specified `dataset_ids` and reference types.  If
    reftypes is None,  all types are returned.

    Returns { dataset_id : { reftype: bestref, ... }, ... }
    """
    try:
        bestrefs = S.get_best_references_by_ids(context, dataset_ids, reftypes, include_headers)
    except Exception as exc:
        raise CrdsLookupError(str(exc)) from exc
    return bestrefs

def get_best_references_by_header_map(context, header_map, reftypes=None):
    """Get best references for header_map = { dataset_id : header, ...}, } and reference types
    where a header is a dictionary of matching parameters.

    If reftypes is None,  all types are returned.

    Returns { dataset_id : { reftype: bestref, ... }, ... }
    """
    try:
        bestrefs_map = S.get_best_references_by_header_map(context, header_map, reftypes)
    except Exception as exc:
        raise CrdsLookupError(str(exc)) from exc
    return bestrefs_map


def get_aui_best_references(date, dataset_ids):
    """Get best references for date and reference types
    where a header is a dictionary of matching parameters.

    Returns { dataset_id : { reftype: bestref, ... }, ... }
    """
    try:
        bestrefs_map = S.get_aui_best_references(date, dataset_ids)
    except Exception as exc:
        raise CrdsLookupError(str(exc)) from exc
    return bestrefs_map

@utils.cached
def get_default_context(observatory=None):
    """Return the name of the latest pipeline mapping in use for processing
    files for `observatory`.
    """
    return str(S.get_default_context(observatory))

@utils.cached
def get_context_by_date(date, observatory=None):
    """Return the name of the first operational context which precedes `date`."""
    return str(S.get_context_by_date(date, observatory))

@utils.cached
def get_server_info():
    """Return a dictionary of critical parameters about the server such as:

    operational_context  - the context in use in the operational pipeline

    edit_context         - the context which was last edited, not
                           necessarily archived or operational yet.

    crds*                - the CRDS package versions on the server.

    This is intended as a single flexible network call which can be used to
    initialize a higher level getreferences() call,  providing information on
    what context, software, and network mode should be used for processing.
    """
    info = _get_server_info()
    info["server"] = get_crds_server()
    # The original CRDS info struct features both "checked" and "unchecked"
    # versions of the download URLs where the unchecked version is a simple
    # static file which has been used exclusively for performance reasons.
    # This should eventually be simplified to a single URL which will nominally
    # be the unchecked version.  Roughly 2020-09-01 is should be possible to
    # simplify the server config and remove the code below from future
    # clients...  while older clients will continue to work with the simplified
    # server.
    if "unchecked" in info.get("reference_url", "UNDEFINED"):
        info["reference_url"] = info["reference_url"]["unchecked"]
    if "unchecked" in info.get("mapping_url", "UNDEFINED"):
        info["mapping_url"] = info["mapping_url"]["unchecked"]
    if "unchecked" in info.get("config_url", "UNDEFINED"):
        info["config_url"] = info["config_url"]["unchecked"]
    if "unchecked" in info.get("pickle_url", "UNDEFINED"):
        info["pickle_url"] = info["pickle_url"]["unchecked"]
    # Add fallback download_metadata for using new client with old servers
    # Put into direct-from-server encoded form decoded later get_download_metadata()
    # but stored in server_config as is.
    if "download_metadata" not in info:
        metadata = get_file_info_map(
            get_default_observatory(), fields=["size", "sha1sum"])
        info["download_metadata"] = proxy.crds_encode(metadata)
    return info

@utils.cached
def get_download_metadata():
    "Defer and cache decoding of download_metadata field of server info."""
    info = get_server_info()
    return proxy.crds_decode(info["download_metadata"])

def _get_server_info():
    """Fetch the server info dict.   If CRDS_CONFIG_URI is set then
    download that URL and load json from the contents.  Otherwise,
    call the CRDS server JSONRPC get_server_info() API.

    Returns  server info dict
    """
    config_uri = config.get_uri("server_config")
    try:
        if config_uri != "none":
            log.verbose(f"Loading config from URI '{config_uri}'.")
            content = utils.get_uri_content(config_uri)
            info = ast.literal_eval(content)
            info["status"] = "uri"
            info["connected"] = False
        else:
            config_uri = f"JSON RPC service at '{get_crds_server()}'"
            info = S.get_server_info()
            log.verbose("Connected to server at", srepr(get_crds_server()))
            info["status"] = "server"
            info["connected"] = True
    except Exception as exc:
        raise CrdsNetworkError(
            f"Failed downloading cache config from: {config_uri}:",
            srepr(exc)) from exc
    return info

get_cached_server_info = get_server_info

def get_server_version():
    """Return the API version of the current CRDS server."""
    info = get_server_info()
    return info["crds_version"]["str"]

def get_dataset_headers_by_id(context, dataset_ids, datasets_since=None):
    """Return { dataset_id : { header } } for `dataset_ids`."""
    context = os.path.basename(context)
    return S.get_dataset_headers_by_id(context, dataset_ids, datasets_since)

def get_dataset_ids(context, instrument, datasets_since=None):
    """Return [ dataset_id, ...] for `instrument`."""
    context = os.path.basename(context)
    return S.get_dataset_ids(context, instrument, datasets_since)

@utils.cached
def get_required_parkeys(context):
    """Return a mapping from instruments to lists of parameter names required to
    compute bestrefs under `context`,  i.e. matching header keys.

    { instrument : [ matching_parkey_name, ... ], ... }
    """
    context = os.path.basename(context)
    return S.get_required_parkeys(context)

def get_dataset_headers_by_instrument(context, instrument, datasets_since=None):
    """return { dataset_id:header, ...} for every `dataset_id` for `instrument`."""
    log.verbose("Dumping datasets for", repr(instrument))
    ids = get_dataset_ids(context, instrument, datasets_since)
    return dict(get_dataset_headers_unlimited(context, ids))

def get_dataset_headers_unlimited(context, ids):
    """Generate (dataset_id, header) for `ids`,  potentially more
    `ids` than can be serviced with a single JSONRPC request.
     If there is a failure fetching parameters for dataset_id,
    `header` will be returned as a string / error message.
    """
    max_ids_per_rpc = get_server_info().get("max_headers_per_rpc", 500)
    for i in range(0, len(ids), max_ids_per_rpc):
        log.verbose("Dumping dataset headers", i , "of", len(ids), verbosity=20)
        id_slice = ids[i : i + max_ids_per_rpc]
        header_slice = get_dataset_headers_by_id(context, id_slice)
        for item in header_slice.items():
            yield item

def get_affected_datasets(observatory, old_context=None, new_context=None):
    """Return a structure describing the ids affected by the last context change."""
    return utils.Struct(S.get_affected_datasets(observatory, old_context, new_context))

def get_context_history(observatory):
    """Fetch the history of context transitions, a list of history era tuples:

     Returns:  [ (start_date, context_name, description), ... ]
    """
    return sorted(tuple(x) for x in S.get_context_history(observatory))

def push_remote_context(observatory, kind, key, context):
    """Upload the specified `context` of type `kind` (e.g. "operational") to the
    server,  informing the server of the actual configuration of the local cache
    for critical systems like pipelines,  not average users.   This lets the server
    display actual versus commanded (Set Context) operational contexts for a pipeline.
    """
    try:
        return S.push_remote_context(observatory, kind, key, context)
    except Exception as exc:
        raise CrdsRemoteContextError(
            "Server error setting pipeline context",
            (observatory, kind, key, context)) from exc

def get_remote_context(observatory, pipeline_name):
    """Get the name of the default context last pushed from `pipeline_name` and
    presumed to be operational.
    """
    try:
        return S.get_remote_context(observatory, pipeline_name)
    except Exception as exc:
        raise CrdsRemoteContextError(
            "Server error resolving context in use by pipeline",
            (observatory, pipeline_name)) from exc

# ==============================================================================

def jpoll_pull_messages(key, since_id=None):
    """Return a list of jpoll json message objects from the channel associated
    with `key` sent after datetime string `since` or since the last pull if
    since is not specified.
    """
    messages = []
    for msg in S.jpoll_pull_messages(key, since_id):
        decoded = utils.Struct(msg)
        decoded.data = html.unescape(decoded.data)
        messages.append(decoded)
    return messages

def jpoll_abort(key):
    """Request that the process writing to jpoll terminate on its next write."""
    return S.jpoll_abort(key)

# ==============================================================================

def get_system_versions(master_version, context=None):
    """Return the versions Struct associated with cal s/w master_version as
    defined by `context` which can be defined as "null", "none", or None to use
    the default context, or with any other valid date based context specifier.
    """
    return utils.Struct(S.get_system_versions(master_version, str(context)))

# ==============================================================================

HARD_DEFAULT_OBS = "jwst"

def get_server_observatory():
    """Return the default observatory according to the server, or None."""
    try:
        pmap = get_default_context()
    except Exception:
        server_obs = None
    else:
        server_obs = observatory_from_string(pmap)
    return server_obs

def get_default_observatory():
    """Based on the environment, cache, and server,  determine the default observatory.

    1. CRDS_OBSERVATORY env var
    2. CRDS_SERVER_URL env var
    3. Observatory(Server default context)
    4. jwst
    """
    obs = config.OBSERVATORY.get()
    if obs != "none":
        return obs
    return observatory_from_string(get_crds_server()) or \
           get_server_observatory() or \
           "jwst"

def observatory_from_string(string):
    """If an observatory name is in `string`, return it,  otherwise return None."""
    for observatory in constants.ALL_OBSERVATORIES:
        if observatory in string:
            return observatory
    return None

# ==============================================================================

def file_progress(activity, name, path, bytes, bytes_so_far, total_bytes, nth_file, total_files):
    """Output progress information for `activity` on file `name` at `path`."""
    return "{activity}  {path!s:<55}  {bytes} bytes  ({nth_file} / {total_files} files) ({bytes_so_far} / {total_bytes} bytes)".format(
        activity=activity,
        path=path,
        bytes=utils.human_format_number(bytes),
        nth_file=nth_file+1,
        total_files=total_files,
        bytes_so_far=utils.human_format_number(bytes_so_far).strip(),
        total_bytes=utils.human_format_number(total_bytes).strip())

# ==============================================================================

class FileCacher:
    """FileCacher gets remote files with simple names into a local cache."""
    def __init__(self, pipeline_context, ignore_cache=False, raise_exceptions=True):
        self.pipeline_context = pipeline_context
        self.observatory = self.observatory_from_context()
        self.ignore_cache = ignore_cache
        self.raise_exceptions = raise_exceptions
        self.info_map = {}

    def get_local_files(self, names):
        """Given a list of basename `mapping_names` which are pertinent to the
        given `pipeline_context`,   cache the mappings locally where they can
        be used by CRDS.
        """
        if isinstance(names, dict):
            names = names.values()
        localpaths = {}

        # Add in GEIS format "conjugate" data files,  .rmaps specify only .rXh
        names2 = names[:]
        for refname in names2:
            if re.match(r"\w+\.r[0-9]h$", refname):
                names.append(refname[:-1]+"d")

        downloads = []
        for name in names:
            localpath = self.locate(name)
            if name.lower() in ["n/a", "undefined"]:
                continue
            if not os.path.exists(localpath):
                downloads.append(name)
            elif self.ignore_cache:
                utils.remove(localpath, observatory=self.observatory)
                downloads.append(name)
                utils.remove(localpath, observatory=self.observatory)
            localpaths[name] = localpath
        if downloads:
            n_bytes = self.download_files(downloads, localpaths)
        else:
            log.verbose("Skipping download for cached files", sorted(names), verbosity=60)
            n_bytes = 0
        return localpaths, len(downloads), n_bytes

    def observatory_from_context(self):
        """Determine the observatory from `pipeline_context`,  based on name if possible."""
        import crds
        for observatory in constants.ALL_OBSERVATORIES:
            if observatory in self.pipeline_context:
                return observatory
        else:
            observatory = crds.get_pickled_mapping(self.pipeline_context).observatory  # reviewed
        return observatory

    def locate(self, name):
        """Return the standard CRDS cache location for file `name`."""
        return config.locate_file(name, observatory=self.observatory)

    def catalog_file_size(self, name):
        """Return the size of file `name` based on the server catalog."""
        return int(self.info_map[os.path.basename(name)]["size"])

    def download_files(self, downloads, localpaths):
        """Serial file-by-file download."""
        download_metadata = get_download_metadata()
        self.info_map = {}
        for filename in downloads:
            self.info_map[filename] = download_metadata.get(filename, "NOT FOUND unknown to server")
        if config.writable_cache_or_verbose("Readonly cache, skipping download of (first 5):", repr(downloads[:5]), verbosity=70):
            bytes_so_far = 0
            total_files = len(downloads)
            total_bytes = get_total_bytes(self.info_map)
            for nth_file, name in enumerate(downloads):
                try:
                    if "NOT FOUND" in self.info_map[name]:
                        raise CrdsDownloadError("file is not known to CRDS server.")
                    bytes, path = self.catalog_file_size(name), localpaths[name]
                    log.info(file_progress("Fetching", name, path, bytes, bytes_so_far, total_bytes, nth_file, total_files))
                    self.download(name, path)
                    bytes_so_far += os.stat(path).st_size
                except Exception as exc:
                    if self.raise_exceptions:
                        raise
                    else:
                        log.error("Failure downloading file", repr(name), ":", str(exc))
            return bytes_so_far
        return 0

    def download(self, name, localpath):
        """Download a single file."""
        # This code is complicated by the desire to blow away failed downloads.  For the specific
        # case of KeyboardInterrupt,  the file needs to be blown away,  but the interrupt should not
        # be re-characterized so it is still un-trapped elsewhere under normal idioms which try *not*
        # to trap KeyboardInterrupt.
        assert not config.get_cache_readonly(), "Readonly cache,  cannot download files " + repr(name)
        try:
            utils.ensure_dir_exists(localpath)
            return proxy.apply_with_retries(self.download_core, name, localpath)
        except Exception as exc:
            self.remove_file(localpath)
            raise CrdsDownloadError(
                "Error fetching data for", srepr(name),
                "at CRDS server", srepr(get_crds_server()),
                "with mode", srepr(config.get_download_mode()),
                ":", str(exc)) from exc
        except:  #  mainly for control-c,  catch it and throw it.
            self.remove_file(localpath)
            raise

    def remove_file(self, localpath):
        """Removes file at `localpath`."""
        log.verbose("Removing file", repr(localpath))
        try:
            os.remove(localpath)
        except Exception:
            log.verbose("Exception during file removal of", repr(localpath))

    def download_core(self, name, localpath):
        """Download and verify file `name` under context `pipeline_context` to `localpath`."""
        if config.get_download_plugin():
            self.plugin_download(name, localpath)
        else:
            generator = self.get_data_http(name)
            self.generator_download(generator, localpath)
        self.verify_file(name, localpath)

    def generator_download(self, generator, localpath):
        """Read all bytes from `generator` until file is downloaded to `localpath.`"""
        with open(localpath, "wb+") as outfile:
            for data in generator:
                outfile.write(data)

    def plugin_download(self, filename, localpath):
        """Run an external program defined by CRDS_DOWNLOAD_PLUGIN to download filename to localpath."""
        url = self.get_url(filename)
        plugin_cmd = config.get_download_plugin()
        plugin_cmd = plugin_cmd.replace("${SOURCE_URL}", url)
        plugin_cmd = plugin_cmd.replace("${OUTPUT_PATH}", localpath)
        plugin_cmd = plugin_cmd.replace("${FILE_SIZE}", self.info_map[filename]["size"])
        plugin_cmd = plugin_cmd.replace("${FILE_SHA1SUM}", self.info_map[filename]["sha1sum"])
        log.verbose("Running download plugin:", repr(plugin_cmd))
        status = os.WEXITSTATUS(os.system(plugin_cmd))
        if status != 0:
            if status == 2:
                raise KeyboardInterrupt("Interrupted plugin.")
            else:
                raise CrdsDownloadError(
                    "Plugin download fail status =", repr(status),
                    "with command:", srepr(plugin_cmd))

    def get_data_http(self, filename):
        """Yield the data returned from `filename` of `pipeline_context` in manageable chunks."""
        url = self.get_url(filename)
        try:
            infile = request.urlopen(url)
            file_size = utils.human_format_number(self.catalog_file_size(filename)).strip()
            stats = utils.TimingStats()
            data = infile.read(config.CRDS_DATA_CHUNK_SIZE)
            while data:
                stats.increment("bytes", len(data))
                status = stats.status("bytes")
                bytes_so_far = " ".join(status[0].split()[:-1])
                log.verbose("Transferred HTTP", repr(url), bytes_so_far, "/", file_size, "bytes at", status[1], verbosity=20)
                yield data
                data = infile.read(config.CRDS_DATA_CHUNK_SIZE)
        except Exception as exc:
            raise CrdsDownloadError(
                "Failed downloading", srepr(filename),
                "from url", srepr(url), ":", str(exc)) from exc
        finally:
            try:
                infile.close()
            except UnboundLocalError:   # maybe the open failed.
                pass

    def get_url(self, filename):
        """Return the URL used to fetch `filename` of `pipeline_context`."""
        return get_flex_uri(filename, self.observatory)

    def verify_file(self, filename, localpath):
        """Check that the size and checksum of downloaded `filename` match the server."""
        remote_info = self.info_map[filename]
        local_length = os.stat(localpath).st_size
        original_length = int(remote_info["size"])
        if original_length != local_length and config.get_length_flag():
            raise CrdsDownloadError(
                "downloaded file size", local_length,
                "does not match server size", original_length)
        if not config.get_checksum_flag():
            log.verbose("Skipping sha1sum with CRDS_DOWNLOAD_CHECKSUMS=False")
        elif remote_info["sha1sum"] not in ["", "none"]:
            original_sha1sum = remote_info["sha1sum"]
            local_sha1sum = utils.checksum(localpath)
            if original_sha1sum != local_sha1sum:
                raise CrdsDownloadError(
                    "downloaded file", srepr(filename),
                    "sha1sum", srepr(local_sha1sum),
                    "does not match server sha1sum", srepr(original_sha1sum))
        else:
            log.verbose("Skipping sha1sum check since server doesn't know it.")

# ==============================================================================

def dump_mappings3(pipeline_context, ignore_cache=False, mappings=None, raise_exceptions=True):
    """Given a `pipeline_context`, determine the closure of CRDS mappings for it and
    cache them on the local file system.

    If mappings is not None,  sync exactly that list of mapping names,  not their closures.

    Returns:   { mapping_basename :   mapping_local_filepath ... }, downloads, bytes
    """
    assert isinstance(ignore_cache, bool)
    if mappings is None:
        mappings = get_mapping_names(pipeline_context)
    mappings = list(reversed(sorted(set(mappings))))
    return FileCacher(pipeline_context, ignore_cache, raise_exceptions).get_local_files(mappings)

def dump_mappings(*args, **keys):
    """See dump_mappings3.

    Returns { mapping_basename :   mapping_local_filepath ... }
    """
    return dump_mappings3(*args, **keys)[0]

def dump_references3(pipeline_context, baserefs=None, ignore_cache=False, raise_exceptions=True):
    """Given a pipeline `pipeline_context` and list of `baserefs` reference
    file basenames,  obtain the set of reference files and cache them on the
    local file system.

    If `basrefs` is None,  sync the closure of references referred to by `pipeline_context`.

    Returns:  { ref_basename :  reference_local_path }, downloads, bytes
    """
    if baserefs is None:
        baserefs = get_reference_names(pipeline_context)
    baserefs = list(baserefs)
    for refname in baserefs:
        if "NOT FOUND" in refname:
            log.verbose("Skipping " + srepr(refname), verbosity=70)
            baserefs.remove(refname)
    baserefs = sorted(set(baserefs))
    return FileCacher(pipeline_context, ignore_cache, raise_exceptions).get_local_files(baserefs)

def dump_references(*args, **keys):
    """See dump_references3.

    Returns { ref_basename :  reference_local_path }
    """
    return dump_references3(*args, **keys)[0]

def dump_files(pipeline_context=None, files=None, ignore_cache=False, raise_exceptions=True):
    """Unified interface to dump any file in `files`, mapping or reference.

    Returns localpaths,  downloads count,  bytes downloaded
    """
    if pipeline_context is None:
        pipeline_context = get_default_context()
    if files is None:
        files = get_mapping_names(pipeline_context)
    mappings = [ os.path.basename(name) for name in files if config.is_mapping(name) ]
    references = [ os.path.basename(name) for name in files if not config.is_mapping(name) ]
    if mappings:
        m_paths, m_downloads, m_bytes = dump_mappings3(
            pipeline_context, mappings=mappings, ignore_cache=ignore_cache, raise_exceptions=raise_exceptions)
    else:
        m_paths, m_downloads, m_bytes = {}, 0, 0
    if references:
        r_paths, r_downloads, r_bytes = dump_references3(
            pipeline_context, baserefs=references, ignore_cache=ignore_cache, raise_exceptions=raise_exceptions)
    else:
        r_paths, r_downloads, r_bytes = {}, 0, 0
    return dict(list(m_paths.items())+list(r_paths.items())), m_downloads + r_downloads, m_bytes + r_bytes

# =====================================================================================================

def cache_best_references(pipeline_context, header, ignore_cache=False, reftypes=None):
    """Given the FITS `header` of a dataset and a `pipeline_context`, determine
    the best set of reference files for processing the dataset,  cache them
    locally,  and return the mapping  { filekind : local_file_path }.
    """
    best_refs = get_best_references(pipeline_context, header, reftypes=reftypes)
    local_paths = cache_references(pipeline_context, best_refs, ignore_cache)
    return local_paths

def cache_references(pipeline_context, bestrefs, ignore_cache=False):
    """Given a CRDS context `pipeline_context` and `bestrefs` dictionary, download missing
    reference files and cache them on the local file system.

    bestrefs    { reference_keyword :  reference_basename }

    Returns:   { reference_keyword :  reference_local_filepath ... }
    """
    wanted = _get_cache_filelist_and_report_errors(bestrefs)

    if config.S3_RETURN_URI:
        localrefs = {name: get_flex_uri(name) for name in wanted}
    else:
        localrefs = FileCacher(pipeline_context, ignore_cache, raise_exceptions=False).get_local_files(wanted)[0]

    refs = _squash_unicode_in_bestrefs(bestrefs, localrefs)

    return refs

def _get_cache_filelist_and_report_errors(bestrefs):
    """Compute the list of files to download based on the `bestrefs` dictionary,
    skimming off and reporting errors, and raising an exception on the last error seen.

    Return the list of files to download,  collapsing complex return types like tuples
    and dictionaries into a list of simple filenames.
    """
    wanted = []
    last_error = None
    for filetype, refname in bestrefs.items():
        if isinstance(refname, tuple):
            wanted.extend(list(refname))
        elif isinstance(refname, dict):
            wanted.extend(refname.values())
        elif isinstance(refname, str):
            if "NOT FOUND" in refname:
                if "n/a" in refname.lower():
                    log.verbose("Reference type", srepr(filetype),
                                "NOT FOUND.  Skipping reference caching/download.", verbosity=70)
                else:
                    last_error = CrdsLookupError(
                        "Error determining best reference for",
                        srepr(filetype), " = ", str(refname)[len("NOT FOUND"):])
                    log.error(str(last_error))
            else:
                log.verbose("Reference type", srepr(filetype), "defined as", srepr(refname))
                wanted.append(refname)
        else:
            last_error = CrdsLookupError(
                "Unhandled bestrefs return value type for", srepr(filetype))
            log.error(str(last_error))
    if last_error is not None:
        raise last_error
    return wanted

def _squash_unicode_in_bestrefs(bestrefs, localrefs):
    """Given bestrefs dictionariesy `bestrefs` and `localrefs`, make sure
    there are no unicode strings anywhere in the keys or complex
    values.
    """
    refs = {}
    for filetype, refname in bestrefs.items():
        if isinstance(refname, tuple):
            refs[str(filetype)] = tuple([str(localrefs[name]) for name in refname])
        elif isinstance(refname, dict):
            refs[str(filetype)] = { name : str(localrefs[name]) for name in refname }
        elif isinstance(refname, str):
            if "NOT FOUND" in refname:
                refs[str(filetype)] = str(refname)
            else:
                refs[str(filetype)] = str(localrefs[refname])
        else:  # can't really get here.
            raise CrdsLookupError(
                "Unhandled bestrefs return value type for", srepr(filetype))
    return refs

# =====================================================================================================

# These functions are deprecated and only work when the full CRDS library is installed,  and only for
# some data file formats (.fits).

def cache_best_references_for_dataset(pipeline_context, dataset,
                                      ignore_cache=False):
    """
    determine the best set of reference files,  cache the references
    locally,  and return the mapping  { filekind : local_file_path }.
    """
    header = get_minimum_header(pipeline_context, dataset, ignore_cache)
    return cache_best_references(pipeline_context, header, ignore_cache)

def get_minimum_header(context, dataset, ignore_cache=False):
    """Given a `dataset` and a `context`,  extract relevant header
    information from the `dataset`.
    """
    import crds
    dump_mappings(context, ignore_cache=ignore_cache)
    ctx = crds.get_pickled_mapping(context)   # reviewed
    return ctx.get_minimum_header(dataset)



"""Generic utility routines used by a variety of modules."""

#  XXXX lots of bogus pylint warnings here,  verify before removing imports.

import sys
import os   # False pylint warning unused import,  verify before removing.
import os.path
import shutil  # False pylint warning unused import,  verify before removing.
import stat
import re
import hashlib
import io
import functools
from collections import Counter, defaultdict
import datetime
import ast
import gc
import json
import warnings

# ===================================================================

# import yaml

# ===================================================================

# from crds import data_file,  import deferred until required

from . import log, config, pysh, exceptions
from .constants import ALL_OBSERVATORIES, INSTRUMENT_KEYWORDS

# from ..client import proxy  # import deferred until required

# from crds.client import api  # deferred import below to simplify basic dependencies

# ===================================================================

def deprecate(deprecated, after_date, alternative):
    """Issue a standard deprecation warning for `deprecated` feature
    (e.g. function name) to be removed `after_date` with suggested
    `alternative` feature as the replacement functionality.
    """
    message = f"Feature {deprecated} has been deprecated " \
              f"and will be removed sometime after {after_date}. " \
              f"Feature {alternative} provides alternative functionality."
    warnings.warn(message, DeprecationWarning)

# ===================================================================

class Struct(dict):
    """A dictionary which supports dotted access to members."""
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as exc:
            raise AttributeError(name) from exc

    def __setattr__(self, name, val):
        self[name] = val

# ===================================================================

def combine_dicts(*post_dicts, **post_vars):
    """Combine positional parameters (dictionaries) and individual
    variables specified by keyword into a single parameter dict.
    """
    vars = dict()
    for pars in post_dicts:
        vars.update(pars)
    vars.update(post_vars)
    return vars

# ===================================================================

def trace_compare(self, other, show_equal=False):
    """Recursively compare object `self` to `other` printing differences
    and optionally equal members.
    """
    log.divider(repr(self) + ":")
    for key, value in self.__dict__.items():
        try:
            ovalue = other.__dict__[key]
        except KeyError:
            print(key, "not present in other")
            continue
        equal = (value == ovalue)
        if show_equal or not equal:
            print(key, equal, value, ovalue)
        if hasattr(value, "_trace_compare"):
            value._trace_compare(ovalue)
    for key in other.__dict__:
        try:
            self.__dict__[key]
        except KeyError:
            print(key, "value not present in self")

# ===================================================================

def flatten(sequence):
    """Given a sequence possibly containing nested lists or tuples,
    flatten the sequence to a single non-nested list of primitives.

    >>> flatten((('META.INSTRUMENT.DETECTOR', 'META.SUBARRAY.NAME'), ('META.OBSERVATION.DATE', 'META.OBSERVATION.TIME')))
    ['META.INSTRUMENT.DETECTOR', 'META.SUBARRAY.NAME', 'META.OBSERVATION.DATE', 'META.OBSERVATION.TIME']
    """
    flattened = []
    for elem in sequence:
        if isinstance(elem, (list, tuple)):
            elem = flatten(elem)
        else:
            elem = [elem]
        flattened.extend(elem)
    return flattened

# ===================================================================

def traced(func):
    """Issue a verbose message showing parameters and possibly return val."""
    @functools.wraps(func)
    def func2(*args, **keys):
        "Decoration wrapper for @trace."
        log.verbose("trace:", func.__name__, args if args else "", keys if keys else "", verbosity=55)
        result = func(*args, **keys)
        log.verbose("trace result:", func.__name__, ":", result, verbosity=55)
        return result
    func2.__name__ = func.__name__ + " [traced]"
    func2.__doc__ = func.__doc__
    func2._traced = True
    return func2

# ===================================================================

def gc_collected(func):
    """Run Python's gc.collect() before and after the decorated function.

    This is pretty slow and may be overkill but was motivated by file
    submission use cases such as "certify" and "insert_references" which
    iterate over large numbers of reference files and,  particularly when
    examining arrays,  may easily exhaust memory,  sometimes leading to
    silent OS or shell level crashes with no traceback.
    """
    @functools.wraps(func)
    def func2(*args, **keys):
        "Decoration wrapper for @gc_collected."
        if config.EXPLICIT_GARBAGE_COLLECTION:
            gc.collect()
        result = None
        try:
            result = func(*args, **keys)
        finally:
            if config.EXPLICIT_GARBAGE_COLLECTION:
                gc.collect()
        return result
    func2.__name__ = func.__name__ + " [gc_collected]"
    func2.__doc__ = func.__doc__
    func2._gc_collected = True
    return func2

# ===================================================================

def cached(func):
    """The cached decorator embeds a dictionary in a function wrapper to
    capture prior results.

    The wrapped function works like the original, except it's faster because it
    fetches results for prior calls from the cache.   The wrapped function has
    extra attributes:

    .cache                      -- { key(parameters): old_result } dictionary
    .uncached(*args, **keys)    -- original unwrapped function
    .readonly(*args, **keys)    -- function variant which uses but doesn't update cache
    .cache_key(*args, **keys)   -- returns tuple used to locate a function call result

    >>> @cached
    ... def sum(x,y):
    ...   print("really doing it.")
    ...   return x+y

    The first call should actually call the unwrapped sum():

    >>> sum(1,2)
    really doing it.
    3

    The second call will return the prior result found in the cache:

    >>> sum(1,2)
    3

    Dump or operate on the cache like this, it's just a dict:

    >>> sum.cache
    {(1, 2): 3}

    By-pass the cache and call the undecorated function like this:

    >>> sum.uncached(1,2)
    really doing it.
    3

    Clear the cache like this:

    >>> sum.cache.clear()
    >>> sum(1,2)
    really doing it.
    3

    A variant of the function which reads but does not update the cache is available.
    After calling the read_only variant the cache is not updated:

    >>> sum.cache.clear()
    >>> sum.readonly(1,2)
    really doing it.
    3
    >>> sum(1,2)
    really doing it.
    3

    However,  the readonly variant will exploit any values in the cache already:

    >>> sum(1,2)
    3
    """
    return CachedFunction(func)

class xcached:
    """Caching decorator which supports auxilliary caching parameters.

    omit_from_key lists keywords or positional indices to be excluded from cache
    key creation:

    >>> @xcached(omit_from_key=[0, "x"])
    ... def sum(x, y, z):
    ...     return x + y + z

    >>> sum(1,2,3)
    6

    >>> sum(2,2,3)
    6

    >>> sum.uncached(2,2,3)
    7

    >>> sum.readonly(2,2,3)
    6
    """
    def __init__(self, *args, **keys):
        """Stash the decorator parameters"""
        self.args = args
        self.keys = keys

    def __call__(self, func):
        """Create a CachedFunction for `func` with extra qualifiers *args and **keys.
        Nomnially executes at import time.
        """
        cached = CachedFunction(func, *self.args, **self.keys)
        cached.__name__ = cached.uncached.__name__ + " [xcached]"
        cached.__doc__ = cached.uncached.__doc__
        return cached

class CachedFunction:
    """Class to support the @cached function decorator.   Called at runtime
    for typical caching version of function.
    """

    cache_set = set()

    def __init__(self, func, omit_from_key=None):
        self.cache = dict()
        self.uncached = func
        self.omit_from_key = [] if omit_from_key is None else omit_from_key
        self.cache_set.add(self)
        self.__doc__ = self.uncached.__doc__
        self.__module__ = self.uncached.__module__
        self.__name__ = self.uncached.__name__ + " [cached]"

    def cache_key(self, *args, **keys):
        """Compute the cache key for the given parameters."""
        args = tuple([ a for (i, a) in enumerate(args) if i not in self.omit_from_key])
        keys = tuple([item for item in keys.items() if item[0] not in self.omit_from_key])
        return args + keys

    def _readonly(self, *args, **keys):
        """Compute (cache_key, func(*args, **keys)).   Do not add to cache."""
        key = self.cache_key(*args, **keys)
        if key in self.cache:
            log.verbose("Cached call", self.uncached.__name__, repr(key), verbosity=80)
            return key, self.cache[key]
        else:
            log.verbose("Uncached call", self.uncached.__name__, repr(key), verbosity=80)
            return key, self.uncached(*args, **keys)

    def readonly(self, *args, **keys):
        """Compute or fetch func(*args, **keys) but do not add to cache.
        Return func(*args, **keys)
        """
        _key, result = self._readonly(*args, **keys)
        return result

    def __call__(self, *args, **keys):
        """Compute or fetch func(*args, **keys).  Add the result to the cache.
        return func(*args, **keys)
        """
        key, result = self._readonly(*args, **keys)
        self.cache[key] = result
        return result

    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)

def clear_function_caches():
    "Clear all the caches created using @utils.cached or @utils.xcached."""
    for cache_func in CachedFunction.cache_set:
        log.verbose("Clearing cache for", repr(cache_func.uncached), verbosity=80)
        cache_func.cache = dict()

def list_cached_functions():
    """List all the functions supporting caching under @utils.cached or @utils.xcached."""
    for cache_func in sorted(CachedFunction.cache_set):
        print(repr(cache_func.uncached))

# ===================================================================

def capture_output(func):
    """Decorate a function with @capture_output to define a CapturedFunction()
    wrapper around it.

    Doesn't currently capture non-python output but could with dup2.

    Decorate any function to wrap it in a CapturedFunction() wrapper:

    >>> @capture_output
    ... def f(x,y):
    ...    print("hi")
    ...    return x + y

    >>> f
    CapturedFunction('f')

    Calling a captured function suppresses its output:

    >>> f(1, 2)
    3

    To call the original undecorated function:

    >>> f.uncaptured(1, 2)
    hi
    3

    If you don't care about the return value,  but want the output:

    >>> f.outputs(1, 2) == 'hi\\n'
    True

    If you need both the return value and captured output:

    >>> f.returns_outputs(1, 2) == (3, 'hi\\n')
    True

    """
    class CapturedFunction:
        """Closure on `func` which supports various forms of output capture."""

        def __repr__(self):
            return "CapturedFunction('%s')" % func.__name__

        def returns_outputs(self, *args, **keys):
            """Call the wrapped function,  capture output,  return (f(), output_from_f)."""
            sys.stdout.flush()
            sys.stderr.flush()
            oldout, olderr = sys.stdout, sys.stderr
            if sys.version_info < (3,0,0):
                out = io.BytesIO()
            else:
                out = io.TextIOWrapper(io.BytesIO())
            sys.stdout, sys.stderr = out, out
            try:
                result = func(*args, **keys)
                out.flush()
                out.seek(0)
                return result, out.read()
            finally:
                sys.stdout, sys.stderr = oldout, olderr

        def suppressed(self, *args, **keys):
            """Call the wrapped function, suppress output,  return f() normally."""
            return self.returns_outputs(*args, **keys)[0]

        def outputs(self, *args, **keys):
            """Call the wrapped function, capture output,  return output_from_f."""
            return self.returns_outputs(*args, **keys)[1]

        def __call__(self, *args, **keys):
            """Call the undecorated function,  capturing and discarding it's output,  returning the result."""
            return self.suppressed(*args, **keys)

        def uncaptured(self, *args, **keys):
            """Call the undecorated function and return the result."""
            return func(*args, **keys)

    return CapturedFunction()

# ===================================================================

def compare_dicts(dict1, dict2):
    """Compare two dictionaries and return a dictionary of added, deleted, and replaced items."""
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        return (dict1, dict2)
    deleted = { key: dict1[key] for key in dict1 if key not in dict2 }
    added = { key: dict2[key] for key in dict2 if key not in dict1 }
    replaced = { key: compare_dicts(dict1[key], dict2[key]) for key in dict1 if key in dict2 and dict1[key] != dict2[key] }
    return dict(deleted=deleted, added=added, replaced=replaced)

# ===================================================================

class TimingStats:
    """Track and compute counts and counts per second."""
    def __init__(self, output=None):
        self.counts = Counter()
        self.started = None
        self.stopped = None
        self.elapsed = None
        self.output = log.info if output is None else output
        self.start()

    def get_stat(self, name):
        """Return the value of statistic `name`."""
        return self.counts[name]

    def increment(self, name, amount=1):
        """Add `amount` to stat count for `name`."""
        self.counts[name] += amount

    def start(self):
        """Start the timing interval."""
        self.started = datetime.datetime.now()
        return self

    def stop(self):
        """Stop the timing interval."""
        self.stopped = datetime.datetime.now()
        self.elapsed = self.stopped - self.started

    def report(self):
        """Output all stats."""
        if not self.stopped:
            self.stop()
        self.msg("STARTED", str(self.started)[:-4])
        self.msg("STOPPED", str(self.stopped)[:-4])
        self.msg("ELAPSED", str(self.elapsed)[:-4])
        for kind in self.counts:
            self.report_stat(kind)

    def report_stat(self, name):
        """Output stats on `name`."""
        count, rate = self.status(name)
        self.msg(count, "at", rate)

    def raw_status(self, name):
        self.stop()
        counts = self.counts[name]
        rate = self.counts[name] / self.elapsed.total_seconds()
        return counts, rate

    def status(self, name):
        """Return human readable (count, rate) for `name`."""
        counts, rate = self.raw_status(name)
        count_str = human_format_number(counts) + " " + name
        rate_str = human_format_number(rate) + " " + name + "-per-second"
        return count_str, rate_str

    def log_status(self, name, intro, total=None):
        """Do log output about stat `name` using `intro` as the descriptive lead-in to
        the stats.
        """
        stat, stat_per_sec = self.raw_status(name)
        if total is not None:
            self.msg(intro, "[",
                     human_format_number(stat), "/",
                     human_format_number(total), name, "]",
                     "[",
                     human_format_number(stat_per_sec), name + "-per-second ]")
        else:
            self.msg(intro,
                     "[", human_format_number(stat), name, "]",
                     "[", human_format_number(stat_per_sec), name + "-per-second ]")

    def msg(self, *args):
        """Format (*args, **keys) using log.format() and call output()."""
        self.output(*args, eol="")

# ===================================================================

def total_size(filepaths):
    """Return the total size of all files in `filepaths` as an integer."""
    return sum([os.stat(filename).st_size for filename in filepaths])

# ===================================================================

def file_size(filepath):
    """Return the size of `filepath` as an integer."""
    return os.stat(filepath).st_size

# ===================================================================

def elapsed_time(func):
    """Decorator to report on elapsed time for a function call."""
    def elapsed_wrapper(*args, **keys):
        stats = TimingStats()
        stats.start()
        result = func(*args, **keys)
        stats.stop()
        stats.msg("Timing for", repr(func.__name__))
        stats.report()
        return result
    elapsed_wrapper.__name__ = func.__name__ + "[elapsed_time]"
    elapsed_wrapper.__doc__ = func.__doc__
    return elapsed_wrapper

# ===================================================================

def human_format_number(number):
    """Reformat `number` by switching to engineering units and dropping to two fractional digits,
    10s of megs for G-scale files.
    """
    convert = [
        (1e12, "T"),
        (1e9 , "G"),
        (1e6 , "M"),
        (1e3 , "K"),
        ]
    for limit, sym in convert:
        if isinstance(number, (float, int)) and number > limit:
            number /= limit
            break
    else:
        sym = ""
    if isinstance(number, int):
        # numstr = "%d" % number
        numstr = "{}".format(number)
    else:
        numstr = "{:0.1f} {}".format(number, sym)
    return "{!s:>7}".format(numstr)

# ===================================================================

def invert_dict(dictionary):
    """Return the functional inverse of a dictionary,  raising an exception
    for values in `d` which map to more than one key producing an undefined
    inverse.
    """
    inverse = {}
    for key, value in dictionary.items():
        if value in inverse and inverse[value] != key:
            raise ValueError("Undefined inverse because of duplicates for " +
                             repr(value) + " of " + repr(key) + " vs. " +
                             repr(inverse[value]))
        inverse[value] = key
    return inverse

# ===================================================================

def evalfile(fname):
    """Evaluate and return the contents of file `fname`,  restricting
    expressions to data literals.
    """
    with open(fname) as sourcefile:
        contents = sourcefile.read()
    return ast.literal_eval(contents)

# ===================================================================

UMASK = 0o002
with log.verbose_warning_on_exception("Failed determining UMASK"):
    UMASK = os.umask(0)
    os.umask(UMASK)

DEFAULT_DIR_PERMS = ~UMASK & 0o777

def create_path(path, mode=DEFAULT_DIR_PERMS):
    """Recursively traverses directory path creating directories as
    needed so that the entire path exists.
    """
    path = os.path.abspath(path)
    if os.path.exists(path):
        return
    current = []
    for part in path.split("/"):
        if not part:
            current.append("/")
            continue
        current.append(str(part))
        subdir = os.path.abspath(os.path.join(*current))
        if not os.path.exists(subdir):
            log.verbose("Creating", repr(subdir), "with permissions %o" % mode)
            os.mkdir(subdir, mode)
            with log.verbose_warning_on_exception(
                    "Failed chmod'ing new directory", repr(subdir), "to %o." % mode):
                os.chmod(subdir, mode)

def ensure_dir_exists(fullpath, mode=DEFAULT_DIR_PERMS):
    """Creates dirs from `fullpath` if they don't already exist.  fullpath
    is assumed to include a filename which will not be created.
    """
    create_path(os.path.dirname(fullpath), mode)

# ===================================================================

def is_writable(filepath, no_exist=True):
    """Interpret the mode bits of `filepath` in terms of the current user and it's groups,
    returning True if any of user, group, or other have write permission on the path.

    If `filepath` doesn't exist,  return `no_exist` if the directory is writable.
    """
    if not os.path.exists(filepath):   # If file doesn't exist,  make sure directory is writable.
        return no_exist and len(os.path.dirname(filepath)) and is_writable(os.path.dirname(filepath))
    stats = os.stat(filepath)
    user_writeable = bool(stats.st_mode & stat.S_IWUSR)
    effective_user_matches =  bool(stats.st_uid == os.geteuid())
    group_writeable = bool(stats.st_mode & stat.S_IWGRP)
    group_matches = bool(stats.st_gid in os.getgroups())
    other_writeable = bool(stats.st_mode & stat.S_IWOTH)
    return bool((user_writeable and effective_user_matches) or
                (group_writeable and group_matches) or
                (other_writeable))

# ===================================================================

def remove(rmpath, observatory):
    """Wipe out directory at 'rmpath' somewhere in cache for `observatory`."""
    if config.writable_cache_or_verbose("Skipped removing", repr(rmpath)):
        with log.error_on_exception("Failed removing", repr(rmpath)):
            abs_path = os.path.abspath(rmpath)
            abs_cache = os.path.abspath(config.get_crds_path())
            abs_config = os.path.abspath(config.get_crds_cfgpath(observatory))
            abs_root_config = os.path.abspath(config.get_crds_root_cfgpath())
            abs_references = os.path.abspath(config.get_crds_refpath(observatory))
            abs_mappings = os.path.abspath(config.get_crds_mappath(observatory))
            abs_pickles = os.path.abspath(config.get_crds_picklepath(observatory))
            assert abs_path.startswith((abs_cache, abs_config, abs_root_config,
                                        abs_references, abs_mappings, abs_pickles)), \
                "remove() only works on files in CRDS cache. not: " + repr(rmpath)
            log.verbose("CACHE removing:", repr(rmpath))
            if os.path.isfile(rmpath):
                os.remove(rmpath)
            else:
                pysh.sh("rm -rf ${rmpath}", raise_on_error=True)

# ===================================================================

def _no_message(*args):
    """Do nothing message handler."""

def copytree(src, dst, symlinks=False, fnc_directory=_no_message,
             fnc_file=_no_message, fnc_symlink=_no_message):
    """Derived from shutil.copytree() example with added function hooks called
    on a per-directory, per-file, and per-symlink basis with (src, dest)
    parameters.  Removes exception trapping since partial copies are useless
    for CRDS.  Cannot handle devices or sockets, only regular files and
    directories.   File stats not preserved.
    """
    os.makedirs(dst)
    for name in os.listdir(src):
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        if symlinks and os.path.islink(srcname):
            linkto = os.readlink(srcname)
            fnc_symlink("Linking", log.srepr(linkto), "to", log.srepr(dstname))
            os.symlink(linkto, dstname)
        elif os.path.isdir(srcname):
            fnc_directory("Copying dir", log.srepr(srcname), "to", log.srepr(dstname))
            copytree(srcname, dstname, symlinks)
        else:
            fnc_file("Coping", log.srepr(srcname), "to", log.srepr(dstname))
            shutil.copy(srcname, dstname)

# ===================================================================

def get_s3_uri_content(s3_uri, mode="text"):
    """Perform a direct read of an AWS S3 URI using the AWS SDK.

    Returns  contents of `s3_uri`.
    """
    from ..client import proxy

    log.verbose(f"Fetching content from URI: '{s3_uri}'")
    return proxy.apply_with_retries(_get_s3_uri_content, s3_uri, mode)


def _get_s3_uri_content(s3_uri, mode):
    bucket_name, key = s3_uri.replace("s3://", "").split("/", 1)
    import boto3
    s3 = boto3.resource("s3")
    obj = s3.Object(bucket_name, key)
    binary = obj.get()["Body"].read()
    if mode == "text":
        text = binary.decode("utf-8")
        return text
    return binary


def get_url_content(url, mode="text"):
    """Return the contents of `url` as a string."""
    from ..client import proxy

    log.verbose(f"Fetching content from URL: '{url}'")
    return proxy.apply_with_retries(_get_url_content, url, mode)


def _get_url_content(url, mode):
    import requests
    r = requests.get(url)
    r.raise_for_status()
    if mode == "text":
        return r.text
    return r.content


def get_uri_content(uri, mode="text"):
    """Reads and returns the contents of the given s3://, https://
    or filename uri.   Reads to memory, intended for small files.
    """
    from ..client import proxy

    if uri.startswith("s3://"):
        return get_s3_uri_content(uri, mode)
    elif uri.startswith(("http://", "https://")):
        return get_url_content(uri, mode)
    else:
        mode = "r" if (mode=="text") else "rb"
        with open(uri, mode) as file:
            return file.read()

# ===================================================================

def checksum(pathname):
    """Return the CRDS hexdigest for file at `pathname`.   See also
    copy_and_checksum() below which must match sha1sum results.
    """
    xsum = hashlib.sha1()
    with open(pathname, "rb") as infile:
        size = 0
        insize = os.stat(pathname).st_size
        while size < insize:
            block = infile.read(config.CRDS_CHECKSUM_BLOCK_SIZE)
            size += len(block)
            xsum.update(block)
    return xsum.hexdigest()

def copy_and_checksum(source, destination):
    """Copy file from `source` path to `destination` path computing
    sha1sum of source during the copy.   This is a *gross* server-side
    optimization to prevent copying 10's of G's of data and then reading
    it back just to compute sha1sum.   This is necessitated because it's
    not possible to take ownership of files you can both read and delete
    but it is possible to make a personally owned copy which replaces the
    original.  See also checksum() which must have matching results.
    """
    xsum = hashlib.sha1()
    with open(source, "rb") as source_file:
        with open(destination, "wb+") as destination_file:
            size = 0
            insize = os.stat(source).st_size
            while size < insize:
                block = source_file.read(config.CRDS_CHECKSUM_BLOCK_SIZE)
                destination_file.write(block)
                size += len(block)
                xsum.update(block)
    return xsum.hexdigest()

def str_checksum(data):
    """Return the CRDS hexdigest for small strings.   Likewise,  must
    match checksum() and copy_and_checksum() above.

    >>> str_checksum("this is a test.")
    '7728f8eb7bf75ec3cc49364861eec852fc814870'

    """
    if not isinstance(data, bytes):
        data = data.encode("utf-8")
    xsum = hashlib.sha1()
    xsum.update(data)
    return xsum.hexdigest()

# ===================================================================

def get_file_properties(observatory, filename):
    """Return instrument,filekind fields associated with filename."""
    path = config.locate_file(filename, observatory)
    try:
        return get_locator_module(observatory).get_file_properties(path)
    except Exception as exc:
        raise exceptions.FilePropertiesError(f"Failed to determine '{observatory}' instrument or reftype for '{filename}' : '{str(exc)}'") from exc

def get_instruments_filekinds(observatory, filepaths):
    """Given a list of filepaths return the mapping of instruments and
    filekinds covered by the files.
    """
    itmapping = defaultdict(set)
    for filepath in filepaths:
        instrument, filekind = get_file_properties(observatory, filepath)
        itmapping[instrument] |= set([filekind])
    return { instr : sorted([filekind for filekind in itmapping[instr]]) for instr in  itmapping}

def organize_files(observatory, files):
    """Given and `observatory` name and list of `files`, return a dictionary
    that partitions `files` into lists based upon common instrument and
    filekind.

    Returns:  { (instrument, filekind) : [ file1, file4, ...], ... }
    """
    organized = defaultdict(list)
    for file_ in files:
        instrument, filekind = get_file_properties(observatory, file_)
        organized[(instrument,filekind)].append(file_)
    return dict(organized)

# ===================================================================

MODULE_PATH_RE = re.compile(r"^crds(_server)?(\.\w{1,64}){0,10}$")

@cached
def get_object(*args):
    """Import the given `dotted_name` and return the object.

    >>> rmap = get_object("crds.rmap")
    >>> fail = get_object("crds.rmap; eval('2+2')")
    Traceback (most recent call last):
    ...
    AssertionError: Invalid dotted name for get_object() : "crds.rmap; eval('2+2')"
    >>> rmap = get_object("crds","rmap")
    """
    dotted_name = ".".join(args)
    assert MODULE_PATH_RE.match(dotted_name), \
        "Invalid dotted name for get_object() : " + repr(dotted_name)
    parts = dotted_name.split(".")
    pkgpath = ".".join(parts[:-1])
    cls = parts[-1]
    namespace = {}
    import_cmd = "from " + pkgpath + " import " + cls
    with log.augment_exception("Error importing", repr(import_cmd)):
        exec(import_cmd, namespace, namespace)
        return namespace[cls]

# ==============================================================================

DONT_CARE_RE = re.compile(r"^" + r"|".join([
    # "-999","-999\.0",
    # "4294966297.0",
    r"-2147483648.0",
    r"\(\)","N/A","NOT APPLICABLE", "NOT_APPLICABLE"]) + "$")

NUMBER_RE = re.compile(r"^([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?|[+-]?[0-9]+\.)$")

def condition_value(value):
    """Condition `value`,  ostensibly taken from a FITS header or CDBS
    reference file table,  such that it is suitable for appearing in or
    matching an rmap MatchSelector key.

    >>> condition_value('ANY')
    '*'

    # >>> condition_value('-999')
    'N/A'
    # >>> condition_value('-999.0')
    'N/A'
    # >> condition_value('4294966297.0')   # -999
    'N/A'

    >>> condition_value('N/A')
    'N/A'
    >>> condition_value('NOT APPLICABLE')
    'N/A'
    >>> condition_value('NOT_APPLICABLE')
    'N/A'

    >> condition_value('')
    'N/A'

    >>> condition_value('4294967295')
    '-1.0'

    >>> condition_value(False)
    'F'
    >>> condition_value(True)
    'T'

    >>> condition_value(1)
    '1.0'
    >>> condition_value('-9')
    '-9.0'
    >>> condition_value('1.0')
    '1.0'
    >>> condition_value('1.')
    '1.0'

    >>> condition_value('foo')
    'FOO'

    >>> condition_value('iref$uaf12559i_drk.fits')
    'IREF$UAF12559I_DRK.FITS'

    >>> condition_value('2013-11-05 15:21:34')
    '2013-11-05 15:21:34'
    """
    value = str(value).strip().upper()
    if NUMBER_RE.match(value):
        value = str(float(value))
    if DONT_CARE_RE.match(value):
        value = "N/A"
    if value == "ANY":
        value = "*"
    elif value == "4294967295.0":
        value = "-1.0"
    elif value in ["T", "TRUE"]:
        value = "T"
    elif value in ["F", "FALSE"]:
        value = "F"
    return value

def condition_header(header, needed_keys=None):
    """Return a dictionary of all `needed_keys` from `header` after passing
    their values through the CRDS value conditioner.
    """
    header = { key.upper():val for (key, val) in header.items() }
    if not needed_keys:
        needed_keys = header.keys()
    else:
        needed_keys = [ key.upper() for key in needed_keys ]
    conditioned = { key:condition_value(header[key]) for key in needed_keys }
    return conditioned

def _eval_keys(keys):
    """Return the replacement mapping from rmap-visible parkeys to eval-able keys.

    >>> _eval_keys(("META.INSTRUMENT.NAME",))
    {'META.INSTRUMENT.NAME': 'META_INSTRUMENT_NAME'}

    """
    evalable_map = {}
    for key in keys:
        replacement = key.replace(".", "_")
        if replacement != key:
            evalable_map[key] = replacement
    return evalable_map

def condition_header_keys(header):
    """Convert a matching parameter header into the form which supports eval(), ie.
    JWST-style header keys.   Nominally for JWST data model style keys.

    >>> from pprint import pprint as pp
    >>> pp(condition_header_keys({"META.INSTRUMENT.NAME": "NIRISS"}))
    {'META.INSTRUMENT.NAME': 'NIRISS', 'META_INSTRUMENT_NAME': 'NIRISS'}

    """
    header = dict(header)
    evalable_map = _eval_keys(header.keys())
    if evalable_map:
        header.update({ evalable_map[key] : header[key] for key in evalable_map })
    return header

def condition_source_code_keys(code, parkeys):
    """Convert source code expressed in terms of parkeys into source code which works
    with the evalable form of the parkey.   Nominally for JWST data model style keys.

    >>> condition_source_code_keys('META.INSTRUMENT.NAME != "MIRI"', ('META.INSTRUMENT.NAME',))
    'META_INSTRUMENT_NAME != "MIRI"'

    """
    evalable_map = _eval_keys(parkeys)
    for key, replacement in evalable_map.items():
        code = code.replace(key, replacement)
    return code

# ==============================================================================

# Since this imports all observatory packages,  better to cache than put in global.
@cached
def observatory_instrument_tuples():
    """Yield all the installed tuples of (observatory, instrument) available."""
    tuples = []
    for obs in ALL_OBSERVATORIES:
        try:
            instruments = get_object("crds." + obs + ".INSTRUMENTS")
        except ImportError:
            continue
        for instr in instruments:
            tuples.append((obs, instr))
    return tuples

def file_to_observatory(filename):
    """Return the observatory corresponding to reference, mapping, or dataset `filename`."""
    basename = os.path.basename(filename).lower()
    for obs in ALL_OBSERVATORIES:
        if basename.startswith(obs + "_") or (("_" + obs + ".") in basename) or (basename == obs + ".pmap"):
            return obs
    else:
        return instrument_to_observatory(file_to_instrument(filename))

@cached
def instrument_to_observatory(instrument):
    """Given the name of an instrument,  return the associated observatory.

    >>> instrument_to_observatory("acs")
    'hst'
    >>> instrument_to_observatory("ACS")
    'hst'
    >>> instrument_to_observatory("miri")
    'jwst'
    >>> instrument_to_observatory("foo")
    Traceback (most recent call last):
    ...
    ValueError: Unknown instrument 'foo'
    """
    instrument = fix_instrument(instrument.lower())
    for (obs, instr) in observatory_instrument_tuples():
        if instrument == instr:
            return obs
    else:
        raise ValueError("Unknown instrument " + repr(instrument))

@cached
def get_locator_module(observatory):
    """Return the observatory specific module for handling naming, file location, etc."""
    return get_object("crds." + observatory + ".locate")

@cached
def get_observatory_package(observatory):
    """Return the base observatory package."""
    return get_object("crds." + observatory)

def file_to_locator(filename):
    """Given reference or dataset `filename`,  return the associated observatory locator module."""
    return instrument_to_locator(file_to_instrument(filename))

def instrument_to_locator(instrument):
    """Given an instrument,  return the locator module associated with the
    observatory associated with the instrument.
    """
    return get_locator_module(instrument_to_observatory(instrument))

def file_to_instrument(filename):
    """Given reference or dataset `filename`,  return the associated instrument.
    A key aspect of this function versus get_file_properties() is that observatory is not known.
    """
    basename = os.path.basename(filename).lower()
    for (_obs, instr) in observatory_instrument_tuples():
        if ("_" + instr + "_" in basename) or basename.startswith(instr + "_"):
            return instr.upper()
    from crds import data_file
    header = data_file.get_unconditioned_header(filename, needed_keys=tuple(INSTRUMENT_KEYWORDS))
    return header_to_instrument(header)

def header_to_instrument(header, default=None):
    """Given reference or dataset `header`, return the associated instrument.

    >>> header_to_instrument({"INSTRUME":"ACS"}, default="UNDEFINED")
    'ACS'

    >>> header_to_instrument({"META.INSTRUMENT.NAME":"MIRI"}, default=None)
    'MIRI'

    >>> header_to_instrument({"FOO":"MIRI"}, default="UNDEFINED")
    'UNDEFINED'
    """
    val = get_any_of(header,  INSTRUMENT_KEYWORDS, default)
    val = fix_instrument(val)
    if val is None:
        raise KeyError("No instrument keyword defined in header.")
    else:
        return val.upper()

def fix_instrument(instr):
    """"Apply instrument fixers to `instr` to replace obsolete synonyms with standard version."""
    return get_all_instrument_fixers().get(instr, instr)

@cached
def get_all_instrument_fixers():
    """Return the dictionary which maps all weird instrument name synonymns to standard names
    for the combination of all obersvatories.
    """
    all_fixers = {}
    for obs in ALL_OBSERVATORIES:
        obs_pkg = get_observatory_package(obs)
        try:
            obs_fixers = obs_pkg.INSTRUMENT_FIXERS
        except AttributeError:
            continue
        assert not (set(all_fixers.keys()) & set(obs_fixers.keys())), \
            "Two observatories are using the same instrument fixer."
        all_fixers.update(obs_fixers)
    return all_fixers

def get_any_of(getter,  possible_keys,  default=None):
    """Search for the value of any of `possible_keys` in `dictionary`,  returning `default` if none are found.

    >>> get_any_of( {"A":1},  ["C","D","A"], "UNDEFINED")
    1
    >>> get_any_of( {"X":1},  ["C","D","A"], "UNDEFINED")
    'UNDEFINED'
    """
    for key in possible_keys:
        val = getter.get(key.upper(), None)
        if val is None:
            val = getter.get(key.lower(), None)
        if val is not None and val not in ["undefined", "UNDEFINED"]:
            return val
    else:
        return default

def header_to_observatory(header):
    """Given reference or dataset `header`,  return the associated observatory.

    >>> header_to_observatory({"META.INSTRUMENT.NAME":"MIRI"})
    'jwst'
    """
    instr = header_to_instrument(header)
    observ = instrument_to_observatory(instr)
    return observ

def header_to_locator(header):
    """Given reference or dataset `header` dict,  return the observatory's locator module."""
    instr = header_to_instrument(header)
    locator = instrument_to_locator(instr)
    return locator

def get_reference_paths(observatory):
    """Return the list of subdirectories involved with storing references of all instruments."""
    pkg = get_observatory_package(observatory)
    locate = get_locator_module(observatory)
    return sorted({locate.locate_dir(instrument) for instrument in pkg.INSTRUMENTS})

# ===================================================================

def param_combinations(key_values):
    """Recursively combine the value lists that appear in a keyword values dict
    into dictionaries describing combinations of simple values.

    More directly, it reduces a dict describing valid keyword values to
    CRDS bestrefs header dicts:

    { valid keyword values dict... } -->
        [ bestrefs_header1, bestrefs_header2, ...]

    or in more detail:

    { keyword1 : [values_for_keyword1, ...], ...}  -->
        [{ keyword1 : simple_value_for_keyword1, ...}, ...]
    """
    if isinstance(key_values, dict):
        key_values = list(key_values.items())
    if key_values:
        combs = []
        key, values = key_values[0]
        for subcomb in param_combinations(key_values[1:]):
            if len(values):
                for value in values:
                    comb = dict(subcomb)
                    comb[key] = value
                    combs.append(comb)
            else:
                combs.append(subcomb)
        return combs
    else:
        return [{}]

def write_combs_json(outpath, combs):
    """Write out a list of header dictionaries in JSON format
    suitable for CRDS bestrefs --load-pickles,  inventing "case"
    names for each combination.
    """
    with open(outpath, "w+") as outfile:
        for i, comb in enumerate(combs):
            case = { "CASE_" + str(i) : comb }
            outfile.write(json.dumps(case) + "\n")

def yaml_pars_to_json_bestrefs(yaml_filename, json_filename=None):
    """Given an input file describing parameter values that need
    to be covered,  convert the coverage specification into
    discrete parameter combinations (nominally bestrefs headers)
    and write them out in JSON format suitable for bestrefs.

    META.INSTRUMENT.NAME:
        [ NIRCAM ]

    META.EXPOSURE.READPATT:
        [DEEP8, DEEP2, MEDIUM8, MEDIUM2, SHALLOW4, SHALLOW2,
        BRIGHT2, BRIGHT1,RAPID]
    """
    import yaml
    if json_filename is None:
        json_filename = os.path.splitext(yaml_filename)[0] + ".json"
    with open(yaml_filename) as yaml_file:
        pars = yaml.safe_load(yaml_file)
    combs = param_combinations(pars)
    write_combs_json(json_filename, combs)

# ===================================================================

def fix_json_strings(source_json):
    """Squash unicode in nested json object `source_json`."""
    if sys.version_info >= (3, 0, 0):
        return source_json
    if isinstance(source_json, dict):
        result = {}
        for key, val in source_json.items():
            result[str(key)] = fix_json_strings(val)
    elif isinstance(source_json, (list, tuple)):
        result = []
        for val in source_json:
            result.append(fix_json_strings(val))
    elif isinstance(source_json, str):
        result = str(source_json)
    else:
        result = source_json
    return result

# These functions should actually be general,  working on both references and
# dataset files.
reference_to_instrument = file_to_instrument
reference_to_locator = file_to_locator

def get_observatory_plugin(plugin_name, observatory=None):
    """Get the mission-specific plugin object `plugin_name` which is typically a
    customization function.

    Can be called from contexts where `observatory` is not known locally but
    won't work correctly in all configurations of CRDS;  works best when
    CRDS_OBSERVATORY is defined or `observatory` appears in CRDS_SERVER_URL.

    Returns locator object named `plugin_name`.
    """
    if observatory is None:
        from crds.client import api
        observatory = api.get_default_observatory()
    locator = get_locator_module(observatory)
    return get_object(f"crds.{observatory}.locate.{plugin_name}")

def test():
    """Run doctests."""
    import doctest
    from crds.core import utils
    return doctest.testmod(utils)

if __name__ == "__main__":
    print(test())



"""This module defines replacement functionality for the CDBS "certify" program
used to check parameter values in .fits reference files.   It verifies that FITS
files define required parameters and that they have legal values.
"""
from crds.core import log, utils
from . import core as core_validators
from . import synphot as synphot_validators


__all__ = [
    "validator",
    "get_validators",
]


_VALIDATOR_MODULES = [
    core_validators,
    synphot_validators
]


def validator(info, context=None):
    """Given TpnInfo object `info`, construct and return a Validator for it."""
    if len(info.values) == 1 and info.values[0].startswith("&"):
        # This block handles &-types like &PEDIGREE and &SYBDATE
        # only called on static TPN infos.
        class_name = "".join([v.capitalize() for v in info.values[0][1:].split("_")]) + "Validator"
        module = next((m for m in _VALIDATOR_MODULES if hasattr(m, class_name)), None)
        if module is None:
            raise ValueError("Unrecognized validator {}, expected class {}".format(info.values[0], class_name))
        rval = getattr(module, class_name)(info, context=context)
    elif info.datatype == "C":
        rval = core_validators.CharacterValidator(info, context=context)
    elif info.datatype == "S":
        rval = core_validators.CaseSensitiveCharacterValidator(info, context=context)
    elif info.datatype == "R":
        rval = core_validators.RealValidator(info, context=context)
    elif info.datatype == "D":
        rval = core_validators.DoubleValidator(info, context=context)
    elif info.datatype == "I":
        rval = core_validators.IntValidator(info, context=context)
    elif info.datatype == "L":
        rval = core_validators.LogicalValidator(info, context=context)
    elif info.datatype == "X":
        if info.keytype == "C":
            rval = core_validators.ColumnExpressionValidator(info, context=context)
        else:
            rval = core_validators.ExpressionValidator(info, context=context)
    else:
        raise ValueError("Unimplemented datatype " + repr(info.datatype))
    return rval


def get_validators(observatory, refpath, context=None):
    """Given `observatory` and a path to a reference file `refpath`,  load the
    corresponding validators that define individual constraints that reference
    should satisfy.
    """
    tpns = _get_reffile_tpninfos(observatory, refpath)
    checkers = [validator(x, context=context) for x in tpns]
    log.verbose("Validators for", repr(refpath), "("+str(len(checkers))+"):\n", log.PP(checkers), verbosity=65)
    return checkers


def _get_reffile_tpninfos(observatory, refpath):
    """Load just the TpnInfo objects for `observatory` and the given `refpath`.
    This entails both "class" TpnInfo's from CDBS as well as TpnInfo objects
    derived from the JWST data models.
    """
    locator = utils.get_locator_module(observatory)
    instrument, filekind = locator.get_file_properties(refpath)
    tpns = list(locator.get_all_tpninfos(instrument, filekind, "tpn"))
    tpns.extend(locator.get_extra_tpninfos(refpath))
    return tpns



import os
from pprint import pprint

# ==================================================================================
import numpy as np

from nose.tools import assert_raises, assert_true

# ==================================================================================

from crds import data_file
from crds.core import utils, log, exceptions
from crds.io import factory, tables

from crds.tests import test_config

# ==================================================================================

def dt_get_fits_type():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/valid.fits")
    'fits'
    >>> test_config.cleanup(old_state)
    """

def dt_get_fits_type_opaque():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/opaque_fts.tmp")
    'fits'
    >>> test_config.cleanup(old_state)
    """

def dt_get_asdf_type():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/valid.asdf")
    'asdf'
    >>> test_config.cleanup(old_state)
    """

def dt_get_asdf_type_opaque():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/opaque_asd.tmp")
    'asdf'
    >>> test_config.cleanup(old_state)
    """

def dt_get_geis_type():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/e1b09593u.r1h")
    'geis'
    >>> test_config.cleanup(old_state)
    """

def dt_get_geis_type_opaque():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/opaque_gs.tmp")
    'geis'
    >>> test_config.cleanup(old_state)
    """

def dt_get_json_type():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/valid.json")
    'json'
    >>> test_config.cleanup(old_state)
    """

def dt_get_json_type_opaque():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/opaque_jsn.tmp")
    'json'
    >>> test_config.cleanup(old_state)
    """

def dt_get_yaml_type():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/valid.yaml")
    'yaml'
    >>> test_config.cleanup(old_state)
    """

def dt_get_yaml_type_opaque():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> factory.get_filetype("data/opaque_yml.tmp")
    'yaml'
    >>> test_config.cleanup(old_state)
    """

def dt_fits_table():
    """
    ----------------------------------------------------------------------------------
    Demo and test the basic API for FITS tables:

    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> FITS_FILE = "data/v8q14451j_idc.fits"
    >>> tables.ntables(FITS_FILE)
    1

    >>> for tab in tables.tables(FITS_FILE):
    ...     print(repr(tab))
    SimpleTable('v8q14451j_idc.fits', 1, colnames=('DETCHIP', 'DIRECTION', 'FILTER1', 'FILTER2', 'XSIZE', 'YSIZE', 'XREF', 'YREF', 'V2REF', 'V3REF', 'SCALE', 'CX10', 'CX11', 'CX20', 'CX21', 'CX22', 'CX30', 'CX31', 'CX32', 'CX33', 'CX40', 'CX41', 'CX42', 'CX43', 'CX44', 'CY10', 'CY11', 'CY20', 'CY21', 'CY22', 'CY30', 'CY31', 'CY32', 'CY33', 'CY40', 'CY41', 'CY42', 'CY43', 'CY44'), nrows=694)

    >>> tab.segment
    1

    >>> tab.rows[0]     # doctest: +ELLIPSIS
    (1, 'FORWARD', 'F475W', 'CLEAR2L', 4096, 2048, ...

    >>> len(tab.rows[0]) == len(tab.colnames)
    True

    >>> tab.colnames[0]
    'DETCHIP'

    >>> tab.columns['DETCHIP'][:1]
    (1,)
    >>> test_config.cleanup(old_state)
    """

def dt_csv_table():
    """
    ----------------------------------------------------------------------------------
    Demo and test the API for non-FITS formats using astropy format guessing:

    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> CSV_FILE = "data/ascii_tab.csv"
    >>> tables.ntables(CSV_FILE)
    1

    >>> for tab in tables.tables(CSV_FILE):
    ...     print(repr(tab))
    SimpleTable('ascii_tab.csv', 1, colnames=('OBSID', 'REDSHIFT', 'X', 'Y', 'OBJECT'), nrows=2)

    >>> tab.segment
    1

    >>> tab.rows[0]    # doctest: +ELLIPSIS
    (3102, 0.32..., 4167, 4085, 'Q1250+568-A')

    >>> tab.colnames[0]
    'OBSID'

    >>> tab.columns['OBSID'][0]
    3102
    >>> test_config.cleanup(old_state)
    """

# ==================================================================================

def dt_asdf_history_no_entries_description():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> header = data_file.get_header("data/niriss_ref_distortion.asdf")
    >>> header["HISTORY"]
    'UNDEFINED'
    >>> test_config.cleanup(old_state)
    """

def dt_asdf_history_no_entries_list():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> header = data_file.get_header("data/jwst_miri_distortion_0022.asdf")
    >>> print(header["HISTORY"])
    2017-06-02 14:29:39 :: DOCUMENT: MIRI-TN-00001-ETH_Iss2-1_Calibrationproduct_MRS_d2c.  New files created from CDP-6 with updated file structure and V2/V3 instead of XAN/YAN
    >>> test_config.cleanup(old_state)
    """

def dt_asdf_history_with_entries():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> header = data_file.get_header("data/jwst_nirspec_ifupost_0004.asdf")
    >>> print(header["HISTORY"])
    2018-04-17 20:18:32 :: New version created from CV3 with updated file structure
    >>> test_config.cleanup(old_state)
    """

def dt_get_array_properties_hdu_name():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> pprint(data_file.get_array_properties("data/x2i1559gl_wcp.fits", "_WCP"))
    {'COLUMN_NAMES': ['OPT_ELEM',
                      'XC_RANGE',
                      'RESWIDTH',
                      'MAX_TIME_DIFF',
                      'STEPSIZE',
                      'XD_RANGE',
                      'BOX'],
     'DATA': None,
     'DATA_TYPE': {'BOX': '>i4',
                   'MAX_TIME_DIFF': '>f8',
                   'OPT_ELEM': '|S8',
                   'RESWIDTH': '>f8',
                   'STEPSIZE': '>i4',
                   'XC_RANGE': '>i4',
                   'XD_RANGE': '>i4'},
     'EXTENSION': 1,
     'KIND': 'TABLE',
     'NAME': '_WCP',
     'SHAPE': (4,)}
    >>> test_config.cleanup(old_state)
    """

def dt_get_array_properties_extension_number1():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> pprint(data_file.get_array_properties("data/x2i1559gl_wcp.fits", "EXT1"))
    {'COLUMN_NAMES': ['OPT_ELEM',
                      'XC_RANGE',
                      'RESWIDTH',
                      'MAX_TIME_DIFF',
                      'STEPSIZE',
                      'XD_RANGE',
                      'BOX'],
     'DATA': None,
     'DATA_TYPE': {'BOX': '>i4',
                   'MAX_TIME_DIFF': '>f8',
                   'OPT_ELEM': '|S8',
                   'RESWIDTH': '>f8',
                   'STEPSIZE': '>i4',
                   'XC_RANGE': '>i4',
                   'XD_RANGE': '>i4'},
     'EXTENSION': 1,
     'KIND': 'TABLE',
     'NAME': '1',
     'SHAPE': (4,)}
    >>> test_config.cleanup(old_state)
    """

def dt_get_array_properties_extension_number2():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> pprint(data_file.get_array_properties("data/x2i1559gl_wcp.fits", "EXTENSION1"))
    {'COLUMN_NAMES': ['OPT_ELEM',
                      'XC_RANGE',
                      'RESWIDTH',
                      'MAX_TIME_DIFF',
                      'STEPSIZE',
                      'XD_RANGE',
                      'BOX'],
     'DATA': None,
     'DATA_TYPE': {'BOX': '>i4',
                   'MAX_TIME_DIFF': '>f8',
                   'OPT_ELEM': '|S8',
                   'RESWIDTH': '>f8',
                   'STEPSIZE': '>i4',
                   'XC_RANGE': '>i4',
                   'XD_RANGE': '>i4'},
     'EXTENSION': 1,
     'KIND': 'TABLE',
     'NAME': '1',
     'SHAPE': (4,)}
    >>> test_config.cleanup(old_state)
    """

def dt_get_array_properties_extension_number1():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> pprint(data_file.get_array_properties("data/x2i1559gl_wcp.fits", "EXT1"))
    {'COLUMN_NAMES': ['OPT_ELEM',
                      'XC_RANGE',
                      'RESWIDTH',
                      'MAX_TIME_DIFF',
                      'STEPSIZE',
                      'XD_RANGE',
                      'BOX'],
     'DATA': None,
     'DATA_TYPE': {'BOX': '>i4',
                   'MAX_TIME_DIFF': '>f8',
                   'OPT_ELEM': '|S8',
                   'RESWIDTH': '>f8',
                   'STEPSIZE': '>i4',
                   'XC_RANGE': '>i4',
                   'XD_RANGE': '>i4'},
     'EXTENSION': 1,
     'KIND': 'TABLE',
     'NAME': '1',
     'SHAPE': (4,)}
    >>> test_config.cleanup(old_state)
    """

def dt_get_array_properties_extension_number2():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> pprint(data_file.get_array_properties("data/x2i1559gl_wcp.fits", "EXTENSION1"))
    {'COLUMN_NAMES': ['OPT_ELEM',
                      'XC_RANGE',
                      'RESWIDTH',
                      'MAX_TIME_DIFF',
                      'STEPSIZE',
                      'XD_RANGE',
                      'BOX'],
     'DATA': None,
     'DATA_TYPE': {'BOX': '>i4',
                   'MAX_TIME_DIFF': '>f8',
                   'OPT_ELEM': '|S8',
                   'RESWIDTH': '>f8',
                   'STEPSIZE': '>i4',
                   'XC_RANGE': '>i4',
                   'XD_RANGE': '>i4'},
     'EXTENSION': 1,
     'KIND': 'TABLE',
     'NAME': '1',
     'SHAPE': (4,)}
    >>> test_config.cleanup(old_state)
    """


def dt_get_array_properties_extension_number3():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> pprint( data_file.get_array_properties("data/x2i1559gl_wcp.fits", 1))
    {'COLUMN_NAMES': ['OPT_ELEM',
                      'XC_RANGE',
                      'RESWIDTH',
                      'MAX_TIME_DIFF',
                      'STEPSIZE',
                      'XD_RANGE',
                      'BOX'],
     'DATA': None,
     'DATA_TYPE': {'BOX': '>i4',
                   'MAX_TIME_DIFF': '>f8',
                   'OPT_ELEM': '|S8',
                   'RESWIDTH': '>f8',
                   'STEPSIZE': '>i4',
                   'XC_RANGE': '>i4',
                   'XD_RANGE': '>i4'},
     'EXTENSION': 1,
     'KIND': 'TABLE',
     'NAME': '1',
     'SHAPE': (4,)}
    >>> test_config.cleanup(old_state)
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> pprint(data_file.get_array_properties("data/x2i1559gl_wcp.fits", "1"))
    {'COLUMN_NAMES': ['OPT_ELEM',
                      'XC_RANGE',
                      'RESWIDTH',
                      'MAX_TIME_DIFF',
                      'STEPSIZE',
                      'XD_RANGE',
                      'BOX'],
     'DATA': None,
     'DATA_TYPE': {'BOX': '>i4',
                   'MAX_TIME_DIFF': '>f8',
                   'OPT_ELEM': '|S8',
                   'RESWIDTH': '>f8',
                   'STEPSIZE': '>i4',
                   'XC_RANGE': '>i4',
                   'XD_RANGE': '>i4'},
     'EXTENSION': 1,
     'KIND': 'TABLE',
     'NAME': '1',
     'SHAPE': (4,)}
    >>> test_config.cleanup(old_state)
    """

def dt_get_array_properties_extension_number3():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> pprint( data_file.get_array_properties("data/x2i1559gl_wcp.fits", "_WCP__1"))
    {'COLUMN_NAMES': ['OPT_ELEM',
                      'XC_RANGE',
                      'RESWIDTH',
                      'MAX_TIME_DIFF',
                      'STEPSIZE',
                      'XD_RANGE',
                      'BOX'],
     'DATA': None,
     'DATA_TYPE': {'BOX': '>i4',
                   'MAX_TIME_DIFF': '>f8',
                   'OPT_ELEM': '|S8',
                   'RESWIDTH': '>f8',
                   'STEPSIZE': '>i4',
                   'XC_RANGE': '>i4',
                   'XD_RANGE': '>i4'},
     'EXTENSION': 1,
     'KIND': 'TABLE',
     'NAME': ('_WCP', 1),
     'SHAPE': (4,)}
    >>> test_config.cleanup(old_state)

    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> pprint(data_file.get_array_properties("data/x2i1559gl_wcp.fits", "1"))
    {'COLUMN_NAMES': ['OPT_ELEM',
                      'XC_RANGE',
                      'RESWIDTH',
                      'MAX_TIME_DIFF',
                      'STEPSIZE',
                      'XD_RANGE',
                      'BOX'],
     'DATA': None,
     'DATA_TYPE': {'BOX': '>i4',
                   'MAX_TIME_DIFF': '>f8',
                   'OPT_ELEM': '|S8',
                   'RESWIDTH': '>f8',
                   'STEPSIZE': '>i4',
                   'XC_RANGE': '>i4',
                   'XD_RANGE': '>i4'},
     'EXTENSION': 1,
     'KIND': 'TABLE',
     'NAME': '1',
     'SHAPE': (4,)}
    >>> test_config.cleanup(old_state)
    """

# ==================================================================================

def main():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_io, tstmod
    return tstmod(test_io)

if __name__ == "__main__":
    print(main())



import sys

from .certify import CertifyScript

def main():
    """Construct and run the Certify script,  return 1 if errors occurred, 0 otherwise."""
    errors = CertifyScript()()
    exit_status = int(errors > 0)  # no errors = 0,  errors = 1
    return exit_status

if __name__ == "__main__":
    sys.exit(CertifyScript()())



# List of all the observatory package names
ALL_OBSERVATORIES = ["hst", "jwst", "roman", "tobs"]

# keywords used to identify instrument from headers
INSTRUMENT_KEYWORDS = ["INSTRUME", "META.INSTRUMENT.NAME",
                       "META_INSTRUMENT_NAME", "INSTRUMENT",
                       "ROMAN.META.INSTRUMENT.NAME",
                       "META.INSTRUMENT.TYPE", "META_INSTRUMENT_TYPE"]
